import contextlib
import logging
import pkgutil
import subprocess
import threading
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import requests.exceptions
import sshtunnel

from sky.utils import command_runner
from sshtunnel import HandlerSSHTunnelForwarderError, SSHTunnelForwarder

from runhouse.rh_config import open_cluster_tunnels, rns_client
from runhouse.rns.folders.folder import Folder
from runhouse.rns.packages.package import Package
from runhouse.rns.resource import Resource
from runhouse.rns.utils.hardware import _current_cluster

from runhouse.servers.http import DEFAULT_SERVER_PORT, HTTPClient

logger = logging.getLogger(__name__)


class Cluster(Resource):
    RESOURCE_TYPE = "cluster"
    REQUEST_TIMEOUT = 5  # seconds

    def __init__(
        self,
        name,
        ips: List[str] = None,
        ssh_creds: Dict = None,
        dryrun=False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        The Runhouse cluster, or system. This is where you can run Functions or access/transfer data
        between. You can BYO (bring-your-own) cluster by providing cluster IP and ssh_creds, or
        this can be an on-demand cluster that is spun up/down through
        `SkyPilot <https://github.com/skypilot-org/skypilot>`_, using your cloud credentials.

        .. note::
            To build a cluster, please use the factory method :func:`cluster`.
        """

        super().__init__(name=name, dryrun=dryrun)

        self.address = ips[0] if ips else None
        self._ssh_creds = ssh_creds
        self.ips = ips
        self._rpc_tunnel = None
        self.client = None

        if not dryrun and self.address:
            self.check_server()
            # OnDemandCluster will start ray itself, but will also set address later, so won't reach here.
            self.start_ray()

    def save_config_to_cluster(self):
        import json

        config = self.config_for_rns
        if "sky_state" in config.keys():
            # a bunch of setup commands that mess up json dump
            del config["sky_state"]
        json_config = f"{json.dumps(config)}"

        self.run(
            [
                f"mkdir -p ~/.rh; touch ~/.rh/cluster_config.yaml; echo '{json_config}' > ~/.rh/cluster_config.yaml"
            ]
        )

    @staticmethod
    def from_config(config: dict, dryrun=False):
        resource_subtype = config.get("resource_subtype")
        if resource_subtype == "Cluster":
            return Cluster(**config, dryrun=dryrun)
        elif resource_subtype == "OnDemandCluster":
            from runhouse.rns.hardware import OnDemandCluster

            return OnDemandCluster(**config, dryrun=dryrun)
        else:
            raise ValueError(f"Unknown cluster type {resource_subtype}")

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        self.save_attrs_to_config(config, ["ips"])
        if self.ips is not None:
            config["ssh_creds"] = self.ssh_creds()
        return config

    def is_up(self) -> bool:
        """Check if the cluster is up.

        Example:
            >>> rh.cluster("rh-cpu").is_up()
        """
        return self.address is not None

    def up_if_not(self):
        """Bring up the cluster if it is not up. No-op if cluster is already up.
        This only applies to on-demand clusters, and has no effect on self-managed clusters.

        Example:
            >>> rh.cluster("rh-cpu").up_if_not()
        """
        if not self.is_up():
            if not hasattr(self, "up"):
                raise NotImplementedError(
                    f"Cluster <{self.name}> does not have an up method."
                )
            self.up()
        return self

    def keep_warm(self):
        logger.info(
            f"cluster.keep_warm will have no effect on self-managed cluster {self.name}."
        )

    def start_ray(self):
        """Start Ray on the cluster.

        Example:
            >>> rh.cluster("rh-cpu").start_ray()
        """
        if self.is_up():
            res = self.run(["ray start --head"], stream_logs=False)
            if res[0] == 0:
                return
            if any(
                line.startswith("ConnectionError: Ray is trying to start at")
                for line in res[0][1].splitlines()
            ):
                # Ray is already started
                return
            # Check if ray is installed
            if "ray" not in self._get_pip_installs(strip_versions=True):
                self.run(
                    ["pip install ray==2.4.0"]
                )  # pin to SkyPilot's Ray requirement
                res = self.run(["ray start --head"])
                if not res[0][0]:
                    raise RuntimeError(
                        f"Failed to start ray on cluster <{self.name}>. "
                        f"Error: {res[0][1]}"
                    )
        else:
            raise ValueError(f"Cluster <{self.name}> is not up.")

    def _get_pip_installs(self, strip_versions=False):
        packages = self.run(["pip freeze"], stream_logs=False)[0][1].splitlines()
        if strip_versions:
            packages = [p.split("==")[0] for p in packages]
        return packages

    def _sync_runhouse_to_cluster(self, _install_url=None, env=None):
        if not self.address:
            raise ValueError(f"No address set for cluster <{self.name}>. Is it up?")
        local_rh_package_path = Path(pkgutil.get_loader("runhouse").path).parent

        # Check if runhouse is installed from source and has setup.py
        if (
            not _install_url
            and local_rh_package_path.parent.name == "runhouse"
            and (local_rh_package_path.parent / "setup.py").exists()
        ):
            # Package is installed in editable mode
            local_rh_package_path = local_rh_package_path.parent
            dest_path = f"~/{local_rh_package_path.name}"

            # temp update rsync filters to exclude docs, when syncing over runhouse folder
            org_rsync_filter = command_runner.RSYNC_FILTER_OPTION
            command_runner.RSYNC_FILTER_OPTION = (
                "--filter='dir-merge,- .gitignore,- docs/'"
            )
            self._rsync(
                source=str(local_rh_package_path),
                dest=dest_path,
                up=True,
                contents=True,
            )
            command_runner.RSYNC_FILTER_OPTION = org_rsync_filter

            rh_install_cmd = "pip install ./runhouse"
        # elif local_rh_package_path.parent.name == 'site-packages':
        else:
            # Package is installed in site-packages
            # status_codes = self.run(['pip install runhouse-nightly==0.0.2.20221202'], stream_logs=True)
            # rh_package = 'runhouse_nightly-0.0.1.dev20221202-py3-none-any.whl'
            # rh_download_cmd = f'curl https://runhouse-package.s3.amazonaws.com/{rh_package} --output {rh_package}'
            if not _install_url:
                import runhouse

                _install_url = f"runhouse=={runhouse.__version__}"
            rh_install_cmd = f"pip install {_install_url}"

        install_cmd = (
            f"{env._activate_cmd} && {rh_install_cmd}" if env else rh_install_cmd
        )
        status_codes = self.run([install_cmd], stream_logs=True)

        if status_codes[0][0] != 0:
            raise ValueError(f"Error installing runhouse on cluster <{self.name}>")

    def install_packages(
        self, reqs: List[Union[Package, str]], env: Union["Env", str] = None
    ):
        """Install the given packages on the cluster.

        Args:
            reqs (List[Package or str): List of packages to install on cluster and env
            env (Env or str): Environment to install package on. If left empty, defaults to base environment.
                (Default: ``None``)

        Example:
            >>> cluster.install_packages(reqs=["accelerate", "diffusers"])
            >>> cluster.install_packages(reqs=["accelerate", "diffusers"], env="my_conda_env")
        """
        self.check_server()
        to_install = []
        for package in reqs:
            if isinstance(package, str):
                pkg_obj = Package.from_string(package, dryrun=False)
            else:
                pkg_obj = package

            if isinstance(pkg_obj, dict):
                pkg_obj = Package.from_config(pkg_obj)
                to_install.append(pkg_obj)
            elif isinstance(pkg_obj.install_target, Folder):
                if not pkg_obj.install_target.system == self:
                    pkg_str = pkg_obj.name or Path(pkg_obj.install_target.path).name
                    logging.info(
                        f"Copying local package {pkg_str} to cluster <{self.name}>"
                    )
                    pkg_obj = pkg_obj.to(self)
                to_install.append(pkg_obj)
            else:
                to_install.append(package)  # Just appending the string!
        logging.info(
            f"Installing packages on cluster {self.name}: "
            f"{[req if isinstance(req, str) else str(req) for req in reqs]}"
        )
        self.client.install(to_install, env)

    def get(self, key: str, default: Any = None, stream_logs: bool = True):
        """Get the result for a given key from the cluster's object store."""
        self.check_server()
        res = self.client.get_object(key, stream_logs=stream_logs)
        return res if res is not None else default

    def get_run(self, run_name: str, folder_path: str = None):
        self.check_server()
        return self.client.get_run_object(run_name, folder_path)

    def add_secrets(self, provider_secrets: dict):
        """Copy secrets from current environment onto the cluster"""
        self.check_server()
        return self.client.add_secrets(provider_secrets)

    def put(self, key: str, obj: Any):
        """Put the given object on the cluster's object store at the given key."""
        self.check_server()
        return self.client.put_object(key, obj)

    def list_keys(self):
        """List all keys in the cluster's object store."""
        self.check_server()
        res = self.client.list_keys()
        return res

    def cancel(self, key: str, force=False):
        """Cancel a given run on cluster by its key."""
        self.check_server()
        return self.client.cancel_runs(key, force=force)

    def cancel_all(self, force=False):
        """Cancel all runs on cluster."""
        self.check_server()
        return self.client.cancel_runs("all", force=force)

    def clear_pins(self, pins: Optional[List[str]] = None):
        """Remove the given pinned items from the cluster. If `pins` is set to ``None``, then
        all pinned objects will be cleared."""
        self.check_server()
        self.client.clear_pins(pins)
        logger.info(f'Clearing pins on cluster {pins or ""}')

    def on_this_cluster(self):
        """Whether this function is being called on the same cluster."""
        return _current_cluster("name") == self.rns_address

    # ----------------- RPC Methods ----------------- #

    def connect_server_client(self, tunnel=True, force_reconnect=False):
        # FYI based on: https://sshtunnel.readthedocs.io/en/latest/#example-1
        # FYI If we ever need to do this from scratch, we can use this example:
        # https://github.com/paramiko/paramiko/blob/main/demos/rforward.py#L74
        if not self.address:
            raise ValueError(f"No address set for cluster <{self.name}>. Is it up?")

        # TODO [DG] figure out how to ping to see if tunnel is already up
        if self._rpc_tunnel and force_reconnect:
            self._rpc_tunnel.close()

        # TODO Check if port is already open instead of refcounting?
        # status = subprocess.run(['nc', '-z', self.address, str(self.grpc_port)], capture_output=True)
        # if not self.check_port(self.address, UnaryClient.DEFAULT_PORT):

        tunnel_refcount = 0
        if self.address in open_cluster_tunnels:
            ssh_tunnel, connected_port, tunnel_refcount = open_cluster_tunnels[
                self.address
            ]
            ssh_tunnel.check_tunnels()
            if ssh_tunnel.tunnel_is_up[ssh_tunnel.local_bind_address]:
                self._rpc_tunnel = ssh_tunnel
        else:
            self._rpc_tunnel, connected_port = self.ssh_tunnel(
                HTTPClient.DEFAULT_PORT,
                remote_port=DEFAULT_SERVER_PORT,
                num_ports_to_try=5,
            )
        open_cluster_tunnels[self.address] = (
            self._rpc_tunnel,
            connected_port,
            tunnel_refcount + 1,
        )

        # Connecting to localhost because it's tunneled into the server at the specified port.
        self.client = HTTPClient(host="127.0.0.1", port=connected_port)

    def check_server(self, restart_server=True):
        if self.name == _current_cluster("name"):
            return

        if not self.address:
            # For OnDemandCluster, this initial check doesn't trigger a sky.status, which is slow.
            # If cluster simply doesn't have an address we likely need to up it.
            if not hasattr(self, "up"):
                raise ValueError(
                    "Cluster must have an ip address (i.e. be up) or have a reup_cluster method "
                    "(e.g. OnDemandCluster)."
                )
            if not self.is_up():
                # If this is a OnDemandCluster, before we up the cluster, run a sky.status to see if the cluster
                # is already up but doesn't have an address assigned yet.
                self.up_if_not()

        if not self.client:
            try:
                self.connect_server_client()
                cluster_config = self.config_for_rns
                if "sky_state" in cluster_config.keys():
                    # a bunch of setup commands that mess up json dump
                    del cluster_config["sky_state"]
                logger.info(f"Checking server {self.name}")
                self.client.check_server(cluster_config=cluster_config)
                logger.info(f"Server {self.name} is up.")
            except (
                requests.exceptions.ConnectionError,
                sshtunnel.BaseSSHTunnelForwarderError,
            ):
                # It's possible that the cluster went down while we were trying to install packages.
                if not self.is_up():
                    logger.info(f"Server {self.name} is down.")
                    self.up_if_not()
                elif restart_server:
                    logger.info(
                        f"Server {self.name} is up, but the HTTP server may not be up."
                    )
                    self.restart_server(resync_rh=False)
                    logger.info(f"Checking server {self.name} again.")
                    self.client.check_server(cluster_config=cluster_config)
                else:
                    raise ValueError(f"Could not connect to cluster <{self.name}>")
        return

    def ssh_tunnel(
        self, local_port, remote_port=None, num_ports_to_try: int = 0
    ) -> Tuple[SSHTunnelForwarder, int]:
        # Debugging cmds (mac):
        # netstat -vanp tcp | grep 5005
        # lsof -i :5005_
        # kill -9 <pid>

        creds: dict = self.ssh_creds()
        connected = False
        ssh_tunnel = None
        while not connected:
            try:
                if local_port > local_port + num_ports_to_try:
                    raise Exception(
                        f"Failed to create SSH tunnel after {num_ports_to_try} attempts"
                    )

                ssh_tunnel = SSHTunnelForwarder(
                    self.address,
                    ssh_username=creds.get("ssh_user"),
                    ssh_pkey=creds.get("ssh_private_key"),
                    local_bind_address=("", local_port),
                    remote_bind_address=("127.0.0.1", remote_port or local_port),
                    set_keepalive=1,
                    # mute_exceptions=True,
                )
                ssh_tunnel.start()
                connected = True
            except HandlerSSHTunnelForwarderError:
                # try connecting with a different port - most likely the issue is the port is already taken
                local_port += 1
                num_ports_to_try -= 1
                pass

        return ssh_tunnel, local_port

    # import paramiko
    # ssh = paramiko.SSHClient()
    # ssh.load_system_host_keys()
    # ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    # from pathlib import Path
    # ssh.connect(self.address,
    #             username=creds['ssh_user'],
    #             key_filename=str(Path(creds['ssh_private_key']).expanduser())
    #             )
    # transport = ssh.get_transport()
    # transport.request_port_forward('', local_port)
    # ssh_tunnel = transport.open_channel("direct-tcpip", ("localhost", local_port),
    #                                     (self.address, remote_port or local_port))
    # if ssh_tunnel.is_active():
    #     connected = True
    #     print(f"SSH tunnel is open to {self.address}:{local_port}")

    def restart_server(
        self,
        _rh_install_url: str = None,
        resync_rh: bool = True,
        restart_ray: bool = False,
    ):
        """Restart the RPC server.

        Args:
            resync_rh (bool): Whether to resync runhouse. (Default: True)
            restart_ray (bool): Whether to restart Ray. (Default: False)

        Example:
            >>> rh.cluster("rh-cpu").restart_server()
        """
        logger.info(f"Restarting HTTP server on {self.name}.")

        # TODO how do we capture errors if this fails?
        if resync_rh:
            self._sync_runhouse_to_cluster(_install_url=_rh_install_url)
        logfile = f"cluster_server_{self.name}.log"
        http_server_cmd = "python -m runhouse.servers.http.http_server"
        kill_proc_cmd = f'pkill -f "{http_server_cmd}"'
        # 2>&1 redirects stderr to stdout
        screen_cmd = (
            f"screen -dm bash -c '{http_server_cmd} |& tee -a ~/.rh/{logfile} 2>&1'"
        )
        cmds = [kill_proc_cmd]
        if restart_ray:
            ray_start_cmd = "ray start --head --port 6379 --autoscaling-config=~/ray_bootstrap_config.yaml"
            # We need to use this instead of ray stop to make sure we don't stop the SkyPilot ray server,
            # which runs on other ports but is required to preserve autostop and correct cluster status.
            kill_ray_cmd = 'pkill -f ".*ray.*6379.*"'
            if self.ips and len(self.ips) > 1:
                raise NotImplementedError(
                    "Starting Ray on a cluster with multiple nodes is not yet supported."
                    "In the meantime, you can simply start the Ray cluster via the following instructions, "
                    "and pass *only* the head node ip to the cluster constructor: \n"
                    "https://docs.ray.io/en/latest/cluster/vms/user-guides/launching-clusters/on-premises.html#manually-set-up-a-ray-cluster"
                )
            cmds.append(kill_ray_cmd)
            cmds.append(ray_start_cmd)
        cmds.append(screen_cmd)

        # If we need different commands for debian or ubuntu, we can use this:
        # Need to get actual provider in case provider == 'cheapest'
        # handle = sky.global_user_state.get_cluster_from_name(self.name)['handle']
        # cloud_provider = str(handle.launched_resources.cloud)
        # ubuntu_kill_proc_cmd = f'fuser -k {UnaryService.DEFAULT_PORT}/tcp'
        # debian_kill_proc_cmd = "kill -9 $(netstat -anp | grep 50052 | grep -o '[0-9]*/' | sed 's+/$++')"
        # f'kill -9 $(lsof -t -i:{UnaryService.DEFAULT_PORT})'
        # kill_proc_at_port_cmd = debian_kill_proc_cmd if cloud_provider == 'GCP' \
        #     else ubuntu_kill_proc_cmd

        status_codes = self.run(
            commands=cmds,
            stream_logs=True,
        )
        # As of 2023-15-May still seems we need this.
        time.sleep(5)
        return status_codes

    @contextlib.contextmanager
    def pause_autostop(self):
        """Context manager to temporarily pause autostop. Mainly for OnDemand clusters, for BYO cluster
        there is no autostop."""
        pass

    def _run_module(
        self,
        relative_path,
        module_name,
        fn_name,
        fn_type,
        resources,
        conda_env,
        env_vars,
        run_name,
        args,
        kwargs,
    ):
        self.check_server()
        return self.client.run_module(
            relative_path,
            module_name,
            fn_name,
            fn_type,
            resources,
            conda_env,
            env_vars,
            run_name,
            args,
            kwargs,
        )

    def is_connected(self):
        """Whether the RPC tunnel is up.

        Example:
            >>> connected = cluster.is_connected()
        """
        return self.client is not None

    def disconnect(self):
        """Disconnect the RPC tunnel.

        Example:
            >>> cluster.disconnect()
        """
        if self._rpc_tunnel:
            self._rpc_tunnel.stop()
        # if self.client:
        #     self.client.shutdown()

    def __getstate__(self):
        """Delete non-serializable elements (e.g. thread locks) before pickling."""
        state = self.__dict__.copy()
        state["client"] = None
        state["_rpc_tunnel"] = None
        return state

    # ----------------- SSH Methods ----------------- #

    def ssh_creds(self):
        """Retrieve SSH credentials."""
        return self._ssh_creds

    def _rsync(self, source: str, dest: str, up: bool, contents: bool = False):
        """
        Sync the contents of the source directory into the destination.

        .. note:
            Ending `source` with a slash will copy the contents of the directory into dest,
            while omitting it will copy the directory itself (adding a directory layer).
        """
        # FYI, could be useful: https://github.com/gchamon/sysrsync
        if contents:
            source = source + "/" if not source.endswith("/") else source
            dest = dest + "/" if not dest.endswith("/") else dest
        ssh_credentials = self.ssh_creds()
        runner = command_runner.SSHCommandRunner(self.address, **ssh_credentials)
        if up:
            runner.run(["mkdir", "-p", dest], stream_logs=False)
        else:
            Path(dest).expanduser().parent.mkdir(parents=True, exist_ok=True)
        runner.rsync(source, dest, up=up, stream_logs=False)

    def ssh(self):
        """SSH into the cluster

        Example:
            >>> rh.cluster("rh-cpu").ssh()
        """
        creds = self.ssh_creds()
        subprocess.run(
            f"ssh {creds['ssh_user']}:{self.address} -i {creds['ssh_private_key']}".split(
                " "
            )
        )

    def _ping(self, timeout=5):
        ssh_call = threading.Thread(target=lambda: self.run(['echo "hello"']))
        ssh_call.start()
        ssh_call.join(timeout=timeout)
        if ssh_call.is_alive():
            raise TimeoutError("SSH call timed out")
        return True

    def run(
        self,
        commands: List[str],
        env: Union["Env", str] = None,
        stream_logs: bool = True,
        port_forward: Optional[int] = None,
        require_outputs: bool = True,
        run_name: Optional[str] = None,
    ) -> list:
        """Run a list of shell commands on the cluster. If `run_name` is provided, the commands will be
        sent over to the cluster before being executed and a Run object will be created.

        Example:
            >>> cpu.run(["pip install numpy"])
            >>> cpu.run(["pip install numpy", env="my_conda_env"])
            >>> cpu.run(["python script.py"], run_name="my_exp")
        """
        # TODO [DG] suspect autostop while running?
        from runhouse import run

        cmd_prefix = ""
        if env:
            if isinstance(env, str):
                from runhouse.rns.envs import Env

                env = Env.from_name(env)
            cmd_prefix = env._run_cmd

        if not run_name:
            # If not creating a Run then just run the commands via SSH and return
            return self._run_commands_with_ssh(
                commands, cmd_prefix, stream_logs, port_forward, require_outputs
            )

        # Create and save the Run locally
        with run(name=run_name, cmds=commands, overwrite=True) as r:
            return_codes = self._run_commands_with_ssh(
                commands, cmd_prefix, stream_logs, port_forward, require_outputs
            )

        # Register the completed Run
        r._register_cmd_run_completion(return_codes)
        logger.info(f"Saved Run to path: {r.folder.path}")
        return return_codes

    def _run_commands_with_ssh(
        self,
        commands: list,
        cmd_prefix: str,
        stream_logs: bool,
        port_forward: int = None,
        require_outputs: bool = True,
    ):
        return_codes = []

        runner = command_runner.SSHCommandRunner(self.address, **self.ssh_creds())
        for command in commands:
            command = f"{cmd_prefix} {command}" if cmd_prefix else command
            logger.info(f"Running command on {self.name}: {command}")
            ret_code = runner.run(
                command,
                require_outputs=require_outputs,
                stream_logs=stream_logs,
                port_forward=port_forward,
            )
            return_codes.append(ret_code)
        return return_codes

    def run_python(
        self,
        commands: List[str],
        env: Union["Env", str] = None,
        stream_logs: bool = True,
        port_forward: Optional[int] = None,
        run_name: Optional[str] = None,
    ):
        """Run a list of python commands on the cluster.

        Example:
            >>> cpu.run_python(['import numpy', 'print(numpy.__version__)'])([""])
        """
        cmd_prefix = "python3 -c"
        if env:
            if isinstance(env, str):
                from runhouse.rns.envs import Env

                env = Env.from_name(env)
            cmd_prefix = f"{env._run_cmd} {cmd_prefix}"
        command_str = "; ".join(commands)
        # If invoking a run as part of the python commands also return the Run object
        return_codes = self.run(
            [f'{cmd_prefix} "{command_str}"'],
            stream_logs=stream_logs,
            port_forward=port_forward,
            run_name=run_name,
        )
        return return_codes

    def sync_secrets(self, providers: Optional[List[str]] = None):
        """Send secrets for the given providers.

        Args:
            providers(List[str] or None): List of providers to send secrets for.
                If `None`, all providers configured in the environment will by sent.

        Example:
            >>> cpu.sync_secrets(providers=["aws", "lambda"])
        """
        from runhouse import Secrets

        Secrets.to(system=self, providers=providers)

    def ipython(self):
        # TODO tunnel into python interpreter in cluster
        pass

    def notebook(
        self,
        persist: bool = False,
        sync_package_on_close: Optional[str] = None,
        port_forward: int = 8888,
    ):
        """Tunnel into and launch notebook from the cluster.

        Example:
            >>> rh.cluster("test-cluster").notebook()
        """
        tunnel, port_fwd = self.ssh_tunnel(local_port=port_forward, num_ports_to_try=10)
        try:
            install_cmd = "pip install jupyterlab"
            jupyter_cmd = f"jupyter lab --port {port_fwd} --no-browser"
            with self.pause_autostop():
                self.run(commands=[install_cmd, jupyter_cmd], stream_logs=True)

        finally:
            if sync_package_on_close:
                if sync_package_on_close == "./":
                    sync_package_on_close = rns_client.locate_working_dir()
                pkg = Package.from_string("local:" + sync_package_on_close)
                self._rsync(source=f"~/{pkg.name}", dest=pkg.local_path, up=False)
            if not persist:
                tunnel.stop()
                kill_jupyter_cmd = f"jupyter notebook stop {port_fwd}"
                self.run(commands=[kill_jupyter_cmd])

    def remove_conda_env(
        self,
        env: Union[str, "CondaEnv"],
    ):
        """Remove conda env from the cluster.

        Example:
            >>> rh.cluster("rh-cpu").remove_conda_env("my_conda_env")
        """
        env_name = env if isinstance(env, str) else env.env_name
        self.run([f"conda env remove -n {env_name}"])



from pathlib import Path
from typing import Dict

import yaml

RESERVED_SYSTEM_NAMES = ["file", "s3", "gs", "azure", "here", "ssh", "sftp"]


def _current_cluster(key="name"):
    """Retrive key value from the current cluster config.
    If key is "config", returns entire config."""
    if Path("~/.rh/cluster_config.yaml").expanduser().exists():
        with open(Path("~/.rh/cluster_config.yaml").expanduser()) as f:
            cluster_config = yaml.safe_load(f)
        if key == "config":
            return cluster_config
        elif key == "cluster_name":
            return cluster_config["name"].rsplit("/", 1)[-1]
        return cluster_config[key]
    else:
        return None


def _get_cluster_from(system, dryrun=False):
    from runhouse.rns.hardware import Cluster

    if isinstance(system, Cluster):
        return system
    if system in RESERVED_SYSTEM_NAMES:
        return system

    if isinstance(system, Dict):
        return Cluster.from_config(system, dryrun)

    if isinstance(system, str):
        try:
            system = Cluster.from_name(name=system, dryrun=dryrun)
        except ValueError:
            # Name not found in RNS. Doing the lookup this way saves us a hop to RNS
            pass

    return system



import codecs
import logging
import re
import sys
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel
from ray import cloudpickle as pickle


logger = logging.getLogger(__name__)

DEFAULT_SERVER_PORT = 50052


class Message(BaseModel):
    data: str


class Args(BaseModel):
    args: Optional[List[Any]]
    kwargs: Optional[Dict[str, Any]]


class Response(BaseModel):
    data: Union[None, str, List[str]]
    error: Optional[str]
    traceback: Optional[str]
    output_type: str


class OutputType:
    EXCEPTION = "exception"
    STDOUT = "stdout"
    STDERR = "stderr"
    SUCCESS = "success"  # No output
    RESULT = "result"
    RESULT_LIST = "result_list"


def pickle_b64(picklable):
    return codecs.encode(pickle.dumps(picklable), "base64").decode()


def b64_unpickle(b64_pickled):
    return pickle.loads(codecs.decode(b64_pickled.encode(), "base64"))


def handle_response(response_data, output_type, err_str):
    if output_type == OutputType.RESULT:
        return b64_unpickle(response_data["data"])
    elif output_type == OutputType.RESULT_LIST:
        # Map, starmap, and repeat return lists of results
        return [b64_unpickle(val) for val in response_data["data"]]
    elif output_type == OutputType.SUCCESS:
        return
    elif output_type == OutputType.EXCEPTION:
        fn_exception = b64_unpickle(response_data["error"])
        fn_traceback = b64_unpickle(response_data["traceback"])
        logger.error(f"{err_str}: {fn_exception}")
        logger.error(f"Traceback: {fn_traceback}")
        raise fn_exception
    elif output_type == OutputType.STDOUT:
        res = response_data["data"]
        # Regex to match tqdm progress bars
        tqdm_regex = re.compile(r"(.+)%\|(.+)\|\s+(.+)/(.+)")
        for line in res:
            if tqdm_regex.match(line):
                # tqdm lines are always preceded by a \n, so we can use \x1b[1A to move the cursor up one line
                # For some reason, doesn't work in PyCharm's console, but works in the terminal
                print("\x1b[1A\r" + line, end="", flush=True)
            else:
                print(line, end="", flush=True)
    elif output_type == OutputType.STDERR:
        res = response_data["data"]
        print(res, file=sys.stderr)



import copy
from pathlib import Path
from typing import Dict, List, Optional, Union

from runhouse.rns.folders import Folder
from runhouse.rns.hardware import Cluster
from runhouse.rns.packages import Package
from runhouse.rns.resource import Resource

from runhouse.rns.utils.hardware import _get_cluster_from


class Env(Resource):
    RESOURCE_TYPE = "env"

    def __init__(
        self,
        name: Optional[str] = None,
        reqs: List[Union[str, Package]] = [],
        setup_cmds: List[str] = None,
        env_vars: Union[Dict, str] = {},
        working_dir: Optional[Union[str, Path]] = None,
        dryrun: bool = True,
        **kwargs,  # We have this here to ignore extra arguments when calling from_config
    ):
        """
        Runhouse Env object.

        .. note::
            To create an Env, please use the factory method :func:`env`.
        """
        super().__init__(name=name, dryrun=dryrun)
        self._reqs = reqs
        self.setup_cmds = setup_cmds
        self.env_vars = env_vars
        self.working_dir = working_dir

    @staticmethod
    def from_config(config: dict, dryrun: bool = False):
        """Create an Env object from a config dict"""
        config["reqs"] = [
            Package.from_config(req, dryrun=True) if isinstance(req, dict) else req
            for req in config.get("reqs", [])
        ]
        config["working_dir"] = (
            Package.from_config(config["working_dir"], dryrun=True)
            if isinstance(config["working_dir"], dict)
            else config["working_dir"]
        )

        resource_subtype = config.get("resource_subtype")
        if resource_subtype == "CondaEnv":
            from runhouse import CondaEnv

            return CondaEnv(**config, dryrun=dryrun)

        return Env(**config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        config.update(
            {
                "reqs": [
                    self._resource_string_for_subconfig(package)
                    for package in self._reqs
                ],
                "setup_cmds": self.setup_cmds,
                "env_vars": self.env_vars,
                "working_dir": self._resource_string_for_subconfig(self.working_dir),
            }
        )
        return config

    @property
    def reqs(self):
        return (self._reqs or []) + ([self.working_dir] if self.working_dir else [])

    @reqs.setter
    def reqs(self, reqs):
        self._reqs = reqs

    def _reqs_to(self, system: Union[str, Cluster], path=None, mount=False):
        """Send self.reqs to the system (cluster or file system)"""
        new_reqs = []
        for req in self.reqs:
            if isinstance(req, str):
                new_req = Package.from_string(req)
                if isinstance(new_req.install_target, Folder):
                    req = new_req

            if isinstance(req, Package) and isinstance(req.install_target, Folder):
                req = (
                    req.to(system, path=path, mount=mount)
                    if isinstance(system, Cluster)
                    else req.to(system, path=path)
                )
            new_reqs.append(req)
        if self.working_dir:
            return new_reqs[:-1], new_reqs[-1]
        return new_reqs, None

    def _setup_env(self, system: Cluster):
        """Install packages and run setup commands on the cluster."""
        if self.reqs:
            system.install_packages(self.reqs)
        if self.setup_cmds:
            system.run(self.setup_cmds)

    def to(self, system: Union[str, Cluster], path=None, mount=False):
        """
        Send environment to the system (Cluster or file system).
        This includes installing packages and running setup commands if system is a cluster.

        Example:
            >>> env = rh.env(reqs=["numpy", "pip"])
            >>> cluster_env = env.to(my_cluster)
            >>> s3_env = env.to("s3", path="s3_bucket/my_env")
        """
        system = _get_cluster_from(system)
        new_env = copy.deepcopy(self)
        new_env.reqs, new_env.working_dir = self._reqs_to(system, path, mount)

        if isinstance(system, Cluster):
            system.check_server()
            new_env._setup_env(system)

        return new_env

    @property
    def _activate_cmd(self):
        return ""

    @property
    def _run_cmd(self):
        return ""



import os
import unittest
from pathlib import Path

import pytest

import runhouse as rh
from ray import cloudpickle as pickle
from runhouse.rh_config import configs

DATA_STORE_BUCKET = "runhouse-folder"
DATA_STORE_PATH = f"/{DATA_STORE_BUCKET}/folder-tests"


def fs_str_rh_fn(folder):
    return folder._fs_str


# ----------------- Run tests -----------------


@pytest.mark.clustertest
def test_from_cluster(cpu_cluster):
    rh.folder(path=str(Path.cwd())).to(cpu_cluster, path="~/my_new_tests_folder")
    tests_folder = rh.folder(system=cpu_cluster, path="~/my_new_tests_folder")
    assert "my_new_tests_folder/test_folder.py" in tests_folder.ls()


@pytest.mark.awstest
def test_create_and_save_data_to_s3_folder():
    data = list(range(50))
    s3_folder = rh.folder(path=DATA_STORE_PATH, system="s3")
    s3_folder.mkdir()
    s3_folder.put({"test_data.py": pickle.dumps(data)}, overwrite=True)

    assert s3_folder.exists_in_system()


@pytest.mark.awstest
def test_read_data_from_existing_s3_folder():
    # Note: Uses folder created above
    s3_folder = rh.folder(path=DATA_STORE_PATH, system="s3")
    fss_file: "fsspec.core.OpenFile" = s3_folder.open(name="test_data.py")
    with fss_file as f:
        data = pickle.load(f)

    assert data == list(range(50))


@pytest.mark.awstest
def test_create_and_delete_folder_from_s3():
    s3_folder = rh.folder(name=DATA_STORE_PATH, system="s3")
    s3_folder.mkdir()

    s3_folder.delete_configs()

    assert not s3_folder.exists_in_system()


@pytest.mark.clustertest
def test_folder_attr_on_cluster(local_folder, cpu_cluster):
    cluster_folder = local_folder.to(cpu_cluster)
    fs_str_cluster = rh.function(fn=fs_str_rh_fn).to(cpu_cluster)
    fs_str = fs_str_cluster(cluster_folder)
    assert fs_str == "file"


@pytest.mark.gcptest
@pytest.mark.awstest
@pytest.mark.clustertest
def test_cluster_tos(cpu_cluster, tmp_path):
    tests_folder = rh.folder(path=str(Path.cwd()))

    tests_folder = tests_folder.to(system=cpu_cluster)
    assert "test_folder.py" in tests_folder.ls(full_paths=False)

    # to local
    local = tests_folder.to("here", path=tmp_path)
    assert "test_folder.py" in local.ls(full_paths=False)

    # to s3
    s3 = tests_folder.to("s3")
    assert "test_folder.py" in s3.ls(full_paths=False)

    s3.rm()

    # to gcs
    gcs = tests_folder.to("gs")
    try:
        assert "test_folder.py" in gcs.ls(full_paths=False)
        gcs.rm()

    except:
        bucket_name = gcs._bucket_name_from_path(gcs.path)
        print(
            f"Permissions to gs bucket ({bucket_name}) may not be fully enabled "
            f"on the cluster {cpu_cluster.name}. For now please manually enable them directly on the cluster. "
            f"See https://cloud.google.com/sdk/gcloud/reference/auth/login"
        )


@pytest.mark.clustertest
def test_local_and_cluster(cpu_cluster, local_folder, tmp_path):
    # Local to cluster
    cluster_folder = local_folder.to(system=cpu_cluster)
    assert "sample_file_0.txt" in cluster_folder.ls(full_paths=False)
    assert cluster_folder._fs_str == "ssh"

    # Cluster to local
    local_from_cluster = cluster_folder.to("here", path=tmp_path)
    assert "sample_file_0.txt" in local_from_cluster.ls(full_paths=False)
    assert local_from_cluster._fs_str == "file"


@pytest.mark.awstest
def test_local_and_s3(local_folder, tmp_path):
    # Local to S3
    s3_folder = local_folder.to(system="s3")
    assert "sample_file_0.txt" in s3_folder.ls(full_paths=False)
    assert s3_folder._fs_str == "s3"

    # S3 to local
    local_from_s3 = s3_folder.to("here", path=tmp_path)
    assert "sample_file_0.txt" in local_from_s3.ls(full_paths=False)
    assert local_from_s3._fs_str == "file"

    s3_folder.rm()


@pytest.mark.gcptest
def test_local_and_gcs(local_folder, tmp_path):
    # Local to GCS
    gcs_folder = local_folder.to(system="gs")
    assert "sample_file_0.txt" in gcs_folder.ls(full_paths=False)
    assert gcs_folder._fs_str == "gs"

    # GCS to local
    local_from_gcs = gcs_folder.to("here", path=tmp_path)
    assert "sample_file_0.txt" in local_from_gcs.ls(full_paths=False)
    assert local_from_gcs._fs_str == "file"

    gcs_folder.rm()


@pytest.mark.awstest
@pytest.mark.clustertest
def test_cluster_and_s3(cpu_cluster, cluster_folder):
    # Cluster to S3
    s3_folder = cluster_folder.to(system="s3")
    assert "sample_file_0.txt" in s3_folder.ls(full_paths=False)
    assert s3_folder._fs_str == "s3"

    # S3 to cluster
    cluster_from_s3 = s3_folder.to(system=cpu_cluster)
    assert "sample_file_0.txt" in cluster_from_s3.ls(full_paths=False)
    assert cluster_from_s3._fs_str == "ssh"

    s3_folder.rm()


@unittest.skip("requires GCS setup")
@pytest.mark.clustertest
def test_cluster_and_gcs(cpu_cluster, cluster_folder):
    # Make sure we have gsutil and gcloud on the cluster - needed for copying the package + authenticating
    cpu_cluster.install_packages(["gsutil"])

    # TODO [JL] might be necessary to install gcloud on the cluster
    # c.run(['sudo snap install google-cloud-cli --classic'])

    # Cluster to GCS
    gcs_folder = cluster_folder.to(system="gs")
    try:
        assert "sample_file_0.txt" in gcs_folder.ls(full_paths=False)
        assert gcs_folder._fs_str == "gs"

        # GCS to cluster
        cluster_from_gcs = gcs_folder.to(system=cpu_cluster)
        assert "sample_file_0.txt" in cluster_from_gcs.ls(full_paths=False)
        assert cluster_from_gcs._fs_str == "ssh"

        gcs_folder.rm()

    except:
        # TODO [JL] automate gcloud access on the cluster for writing to GCS bucket
        bucket_name = gcs_folder._bucket_name_from_path(gcs_folder.path)
        raise PermissionError(
            f"Permissions to gs bucket ({bucket_name}) may not be fully enabled "
            f"on the cluster {cpu_cluster.name}. For now please manually enable them directly on the cluster. "
            f"See https://cloud.google.com/sdk/gcloud/reference/auth/login"
        )


@pytest.mark.awstest
def test_s3_and_s3(local_folder, s3_folder):
    # from one s3 folder to another s3 folder
    new_s3_folder = s3_folder.to(system="s3")
    assert "sample_file_0.txt" in new_s3_folder.ls(full_paths=False)

    new_s3_folder.rm()


@pytest.mark.gcptest
def test_gcs_and_gcs(gcs_folder):
    # from one gcs folder to another gcs folder
    new_gcs_folder = gcs_folder.to(system="gs")
    assert "sample_file_0.txt" in new_gcs_folder.ls(full_paths=False)

    new_gcs_folder.rm()


@pytest.mark.gcptest
@pytest.mark.awstest
@unittest.skip("Doesn't work properly as only full-bucket copy is supported")
def test_s3_and_gcs(s3_folder):
    # *** NOTE: transfers between providers are only supported at the bucket level at the moment (not directory) ***

    s3_folder_to_gcs = s3_folder.to(system="gs")
    assert "sample_file_0.txt" in s3_folder_to_gcs.ls(full_paths=False)
    assert s3_folder_to_gcs._fs_str == "gs"

    s3_folder_to_gcs.rm()


@pytest.mark.gcptest
@pytest.mark.awstest
@unittest.skip("Doesn't work properly as only full-bucket copy is supported")
def test_gcs_and_s3(local_folder, gcs_folder):
    # *** NOTE: transfers between providers are only supported at the bucket level at the moment (not directory) ***

    gcs_folder_to_s3 = gcs_folder.to(system="s3")
    assert "sample_file_0.txt" in gcs_folder_to_s3.ls(full_paths=False)
    assert gcs_folder_to_s3._fs_str == "s3"

    gcs_folder_to_s3.rm()


@pytest.mark.awstest
def test_s3_folder_uploads_and_downloads(local_folder, tmp_path):
    # NOTE: you can also specify a specific path like this:
    # test_folder = rh.folder(path='/runhouse/my-folder', system='s3')

    s3_folder = rh.folder(system="s3")
    s3_folder._upload(src=local_folder.path)

    assert s3_folder.exists_in_system()
    assert "sample_file_0.txt" in s3_folder.ls(full_paths=False)

    downloaded_path_folder = tmp_path / "downloaded_s3"
    s3_folder._download(dest=downloaded_path_folder)

    assert downloaded_path_folder.exists()
    assert "sample_file_0.txt" in rh.folder(path=downloaded_path_folder).ls(
        full_paths=False
    )

    # remove folder in s3
    s3_folder.rm()
    assert not s3_folder.exists_in_system()


@pytest.mark.clustertest
def test_cluster_and_cluster(cpu_cluster, byo_cpu, local_folder):
    # Upload sky secrets to cluster - required when syncing over the folder from c1 to c2
    cpu_cluster.sync_secrets(providers=["sky"])

    cluster_folder_1 = local_folder.to(system=cpu_cluster)
    assert "sample_file_0.txt" in cluster_folder_1.ls(full_paths=False)

    # Cluster 1 to cluster 2
    cluster_folder_2 = cluster_folder_1.to(system=byo_cpu, path=cluster_folder_1.path)
    assert "sample_file_0.txt" in cluster_folder_2.ls(full_paths=False)


@pytest.mark.awstest
@pytest.mark.rnstest
def test_s3_sharing(s3_folder):
    token = os.getenv("TEST_TOKEN") or configs.get("token")
    headers = {"Authorization": f"Bearer {token}"}

    assert (
        token
    ), "No token provided. Either set `TEST_TOKEN` env variable or set `token` in the .rh config file"

    # Login to ensure the default folder / username are saved down correctly
    rh.login(token=token, download_config=True, interactive=False)

    s3_folder.save("@/my-s3-shared-folder")
    s3_folder.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="read",
        notify_users=False,
        headers=headers,
    )

    my_folder = rh.folder(name="@/my-s3-shared-folder")
    assert my_folder.ls() == s3_folder.ls()


def test_github_folder(tmp_path):
    gh_folder = rh.folder(
        path="/", system="github", data_config={"org": "pytorch", "repo": "rfcs"}
    )
    assert gh_folder.ls()

    gh_to_local = gh_folder.to(
        system="file", path=tmp_path / "torchrfcs", data_config={}
    )
    assert (tmp_path / "torchrfcs").exists()
    assert "RFC-0000-template.md" in gh_to_local.ls(full_paths=False)


if __name__ == "__main__":
    unittest.main()



import copy
import inspect
import json
import logging
import os
import re
import time
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import requests

from runhouse import rh_config
from runhouse.rns.api_utils.resource_access import ResourceAccess
from runhouse.rns.api_utils.utils import load_resp_content, read_resp_data
from runhouse.rns.envs import CondaEnv, Env
from runhouse.rns.hardware import Cluster
from runhouse.rns.packages import git_package, Package

from runhouse.rns.resource import Resource
from runhouse.rns.run_module_utils import call_fn_by_type

from runhouse.rns.utils.env import _env_vars_from_file, _get_env_from
from runhouse.rns.utils.hardware import _get_cluster_from

logger = logging.getLogger(__name__)


class Function(Resource):
    RESOURCE_TYPE = "function"
    DEFAULT_ACCESS = "write"

    def __init__(
        self,
        fn_pointers: Optional[Tuple] = None,
        system: Optional[Cluster] = None,
        name: Optional[str] = None,
        env: Optional[Env] = None,
        dryrun: bool = False,
        access: Optional[str] = None,
        resources: Optional[dict] = None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Function object. It comprises of the entrypoint, system/cluster,
        and dependencies necessary to run the service.

        .. note::
                To create a Function, please use the factory method :func:`function`.
        """
        self.fn_pointers = fn_pointers
        self.system = system
        self.env = env
        self.access = access or self.DEFAULT_ACCESS
        self.dryrun = dryrun
        self.resources = resources or {}
        super().__init__(name=name, dryrun=dryrun)

        if not self.dryrun:
            self = self.to(self.system, env=self.env)

    # ----------------- Constructor helper methods -----------------

    @staticmethod
    def from_config(config: dict, dryrun: bool = False):
        """Create a Function object from a config dictionary."""
        if isinstance(config["system"], dict):
            config["system"] = Cluster.from_config(config["system"], dryrun=dryrun)
        if isinstance(config["env"], dict):
            config["env"] = Env.from_config(config["env"], dryrun=dryrun)

        return Function(**config, dryrun=dryrun)

    @classmethod
    def _check_for_child_configs(cls, config):
        """Overload by child resources to load any resources they hold internally."""
        system = config["system"]
        if isinstance(system, str):
            config["system"] = rh_config.rns_client.load_config(name=system)
            # if the system is set to a cluster
            if not config["system"]:
                raise Exception(f"No cluster config saved for {system}")

        config["env"] = _get_env_from(config["env"])
        return config

    def to(
        self,
        system: Union[str, Cluster] = None,
        env: Union[List[str], Env] = [],
        # Variables below are deprecated
        reqs: Optional[List[str]] = None,
        setup_cmds: Optional[List[str]] = [],
    ):
        """
        Set up a Function and Env on the given system.

        See the args of the factory method :func:`function` for more information.

        Example:
            >>> rh.function(fn=local_fn).to(gpu_cluster)
            >>> rh.function(fn=local_fn).to(system=gpu_cluster, env=my_conda_env)
        """
        if setup_cmds:
            warnings.warn(
                "``setup_cmds`` argument has been deprecated. "
                "Please pass in setup commands to the ``Env`` class corresponding to the function instead."
            )

        # to retain backwards compatibility
        if reqs or setup_cmds:
            warnings.warn(
                "``reqs`` and ``setup_cmds`` arguments has been deprecated. Please use ``env`` instead."
            )
            env = Env(reqs=reqs, setup_cmds=setup_cmds)
        elif env and isinstance(env, List):
            env = Env(reqs=env, setup_cmds=setup_cmds)
        else:
            env = env or self.env or Env()
            env = _get_env_from(env)

        if (
            self.dryrun
            or not (system or self.system)
            or self.access not in ["write", "read"]
        ):
            # don't move the function to a system
            self.env = env
            return self

        # We need to backup the system here so the __getstate__ method of the cluster
        # doesn't wipe the client of this function's cluster when deepcopy copies it.
        hw_backup = self.system
        self.system = None
        new_function = copy.deepcopy(self)
        self.system = hw_backup

        if system:
            if isinstance(system, str):
                system = Cluster.from_name(system, dryrun=self.dryrun)
            new_function.system = system
        else:
            new_function.system = self.system

        logging.info("Setting up Function on cluster.")
        # To up cluster in case it's not yet up
        new_function.system.check_server()
        new_env = env.to(new_function.system)
        logging.info("Function setup complete.")
        new_function.env = new_env

        return new_function

    @staticmethod
    def _extract_fn_paths(raw_fn: Callable, reqs: List[str]):
        """Get the path to the module, module name, and function name to be able to import it on the server"""
        if not isinstance(raw_fn, Callable):
            raise TypeError(
                f"Invalid fn for Function, expected Callable but received {type(raw_fn)}"
            )
        # Background on all these dunders: https://docs.python.org/3/reference/import.html
        module = inspect.getmodule(raw_fn)

        # Need to resolve in case just filename is given
        module_path = (
            str(Path(inspect.getfile(module)).resolve())
            if hasattr(module, "__file__")
            else None
        )

        # TODO better way of detecting if in a notebook or interactive Python env
        if (
            not module_path
            or module_path.endswith("ipynb")
            or raw_fn.__name__ == "<lambda>"
        ):
            # The only time __file__ wouldn't be present is if the function is defined in an interactive
            # interpreter or a notebook. We can't import on the server in that case, so we need to cloudpickle
            # the fn to send it over. The __call__ function will serialize the function if we return it this way.
            # This is a short-term hack.
            # return None, "notebook", raw_fn.__name__
            root_path = os.getcwd()
            module_name = "notebook"
            fn_name = raw_fn.__name__
        else:
            root_path = os.path.dirname(module_path)
            module_name = inspect.getmodulename(module_path)
            # TODO __qualname__ doesn't work when fn is aliased funnily, like torch.sum
            fn_name = getattr(raw_fn, "__qualname__", raw_fn.__name__)

            # Adapted from https://github.com/modal-labs/modal-client/blob/main/modal/_function_utils.py#L94
            if getattr(module, "__package__", None):
                module_path = os.path.abspath(module.__file__)
                package_paths = [
                    os.path.abspath(p) for p in __import__(module.__package__).__path__
                ]
                base_dirs = [
                    base_dir
                    for base_dir in package_paths
                    if os.path.commonpath((base_dir, module_path)) == base_dir
                ]

                if len(base_dirs) != 1:
                    logger.info(f"Module files: {module_path}")
                    logger.info(f"Package paths: {package_paths}")
                    logger.info(f"Base dirs: {base_dirs}")
                    raise Exception("Wasn't able to find the package directory!")
                root_path = os.path.dirname(base_dirs[0])
                module_name = module.__spec__.name

        remote_import_path = None
        for req in reqs:
            local_path = None
            if (
                isinstance(req, Package)
                and not isinstance(req.install_target, str)
                and req.install_target.is_local()
            ):
                local_path = Path(req.install_target.local_path)
            elif isinstance(req, str):
                if req.split(":")[0] in ["local", "reqs", "pip"]:
                    req = req.split(":")[1]

                if Path(req).expanduser().resolve().exists():
                    # Relative paths are relative to the working directory in Folders/Packages!
                    local_path = (
                        Path(req).expanduser()
                        if Path(req).expanduser().is_absolute()
                        else Path(rh_config.rns_client.locate_working_dir()) / req
                    )

            if local_path:
                try:
                    # Module path relative to package
                    remote_import_path = str(
                        local_path.name / Path(root_path).relative_to(local_path)
                    )
                    break
                except ValueError:  # Not a subdirectory
                    pass

        return remote_import_path, module_name, fn_name

    # Found in python decorator logic, maybe use
    # func_name = getattr(f, '__qualname__', f.__name__)
    # module_name = getattr(f, '__module__', '')
    # if module_name:
    #     full_name = f'{module_name}.{func_name}'
    # else:
    #     full_name = func_name

    # ----------------- Function call methods -----------------

    def __call__(self, *args, stream_logs=True, run_name: str = None, **kwargs) -> Any:
        """Call the function on its system

        Args:
             *args: Optional args for the Function
             stream_logs (bool): Whether to stream the logs from the Function's execution.
                Defaults to ``True``.
             run_name (Optional[str]): Name of the Run to create. If provided, a Run will be created
                for this function call, which will be executed synchronously on the cluster before returning its result
             **kwargs: Optional kwargs for the Function

        Returns:
            The Function's return value
        """

        fn_type = "call"
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            run_locally = not self.system or self.system.on_this_cluster()
            if not run_locally:
                start = time.time()
                run_obj = self.run(*args, run_name=run_name, **kwargs)
                res = self.system.get(run_obj.name, stream_logs=stream_logs)
                end = time.time()
                logging.info(
                    f"Time to call remote function: {round(end - start, 2)} seconds"
                )
                return res
            else:
                [relative_path, module_name, fn_name] = self.fn_pointers
                conda_env = (
                    self.env.env_name
                    if self.env and isinstance(self.env, CondaEnv)
                    else None
                )
                env_vars = self.env.env_vars
                # If we're on this cluster, don't pickle the result before passing back.
                # We need to pickle before passing back in most cases because the env in
                # which the function executes may have a different set of packages than the
                # server, so when Ray passes a result back into the server it will may fail to
                # unpickle. We assume the user's client has the necessary packages to unpickle
                # their own result.

                return call_fn_by_type(
                    fn_type=fn_type,
                    fn_name=fn_name,
                    relative_path=relative_path,
                    module_name=module_name,
                    resources=self.resources,
                    conda_env=conda_env,
                    env_vars=env_vars,
                    run_name=run_name,
                    args=args,
                    kwargs=kwargs,
                    serialize_res=False,
                )
        else:
            # run the function via http path - user only needs Proxy access
            if self.access != ResourceAccess.PROXY:
                raise RuntimeError("Running http path requires proxy access")
            if not rh_config.rns_client.token:
                raise ValueError(
                    "Token must be saved in the local .rh config in order to use an http path"
                )
            http_url = self.http_url()
            logger.info(f"Running {self.name} via http path: {http_url}")
            resp = requests.post(
                http_url,
                data=json.dumps({"args": args, "kwargs": kwargs}),
                headers=rh_config.rns_client.request_headers,
            )
            if resp.status_code != 200:
                raise Exception(
                    f"Failed to run Function endpoint: {load_resp_content(resp)}"
                )

            return read_resp_data(resp)

    def repeat(self, num_repeats: int, *args, **kwargs):
        """Repeat the Function call multiple times.

        Args:
            num_repeats (int): Number of times to repeat the Function call.
            *args: Positional arguments to pass to the Function
            **kwargs: Keyword arguments to pass to the Function

        Example:
            >>> remote_fn = rh.function(local_fn).to(gpu)
            >>> remote_fn.repeat(num_repeats=5)
        """
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return self._call_fn_with_ssh_access(
                fn_type="repeat", args=[num_repeats, args], kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Function.repeat only works with Write or Read access, not Proxy access"
            )

    def map(self, *args, **kwargs):
        """Map a function over a list of arguments.

        Example:
            >>> def local_sum(arg1, arg2, arg3):
            >>>     return arg1 + arg2 + arg3
            >>>
            >>> remote_fn = rh.function(local_fn).to(gpu)
            >>> remote_fn.map([1, 2], [1, 4], [2, 3])
            >>> # output: [4, 9]

        """
        arg_list = zip(*args)
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return self._call_fn_with_ssh_access(
                fn_type="starmap", args=arg_list, kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Function.map only works with Write or Read access, not Proxy access"
            )

    def starmap(self, args_lists, **kwargs):
        """Like :func:`map` except that the elements of the iterable are expected to be iterables
        that are unpacked as arguments. An iterable of [(1,2), (3, 4)] results in [func(1,2), func(3,4)].

        Example:
            >>> arg_list = [(1,2), (3, 4)]
            >>> # runs the function twice, once with args (1, 2) and once with args (3, 4)
            >>> remote_fn.starmap(arg_list)
        """
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return self._call_fn_with_ssh_access(
                fn_type="starmap", args=args_lists, kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Function.starmap only works with Write or Read access, not Proxy access"
            )

    def enqueue(self, resources: Optional[Dict] = None, *args, **kwargs):
        """
        Enqueue a Function call to be run later. This ensures a function call doesn’t run simultaneously with other
        calls, but will wait until the execution completes.

        Example:
            >>> # This will run the functions sequentially
            >>> [remote_fn.enqueue() for _ in range(3)]
        """
        # Add resources one-off without setting as a Function param
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return self._call_fn_with_ssh_access(
                fn_type="queue",
                resources=resources or self.resources,
                args=args,
                kwargs=kwargs,
            )
        else:
            raise NotImplementedError(
                "Function.enqueue only works with Write or Read access, not Proxy access"
            )

    def remote(self, *args, **kwargs):
        warnings.warn("`remote()` is deprecated, use `run()` instead")
        run_obj = self.run(*args, **kwargs)
        return run_obj.name

    def run(self, *args, run_name: str = None, **kwargs):
        """Run async remote call on cluster.

        Args:
            *args: Optional args for the Function
            run_name (Optional[str]): Name of the Run to create. If not provided, a name will automatically
             be generated.
             **kwargs: Optional kwargs for the Function
        Returns:
            Run: Run object
        Example:
            >>> remote_fn = rh.function(local_fn).to(gpu)
            >>> remote_fn.run(arg1, arg2, run_name="my_async_run")
        """
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            from runhouse import Run

            # Use the run_name if provided, otherwise create a new one using the Function's name
            run_name = run_name or Run._create_new_run_name(self.name)

            run_obj = self._call_fn_with_ssh_access(
                fn_type="remote", run_name=run_name, args=args, kwargs=kwargs
            )

            logger.info(f"Submitted remote call to cluster for {run_name}")
            return run_obj
        else:
            raise NotImplementedError(
                "Function.remote only works with Write or Read access, not Proxy access"
            )

    def get(self, run_key):
        """Get the result of a Function call that was submitted as async using `remote`.

        Args:
            run_key: A single or list of runhouse run_key strings returned by a Function.remote() call. The ObjectRefs
                must be from the cluster that this Function is running on.

        Example:
            >>> remote_fn = rh.function(local_fn).to(gpu)
            >>> remote_fn_run = remote_fn.run()
            >>> remote_fn.get(remote_fn_run.name)
        """
        return self.system.get(run_key)

    def _call_fn_with_ssh_access(
        self, fn_type, resources=None, run_name=None, args=None, kwargs=None
    ):
        # https://docs.ray.io/en/latest/ray-core/tasks/patterns/map-reduce.html
        # return ray.get([map.remote(i, map_func) for i in replicas])
        # TODO allow specifying resources per worker for map
        # TODO [DG] check whether we're on the cluster and if so, just call the function directly via the
        # helper function currently in UnaryServer
        resources = (
            resources or self.resources
        )  # Allow for passing in one-off resources for this specific call
        name = self.name or "anonymous function"
        if self.fn_pointers is None:
            raise RuntimeError(f"No fn pointers saved for {name}")

        [relative_path, module_name, fn_name] = self.fn_pointers

        name = self.name or fn_name or "anonymous function"
        logger.info(f"Running {name} via HTTP")
        env_name = (
            self.env.env_name if (self.env and isinstance(self.env, CondaEnv)) else None
        )
        env_vars = self.env.env_vars if self.env else {}
        if not isinstance(env_vars, dict):
            env_vars = _env_vars_from_file(env_vars)

        res = self.system._run_module(
            relative_path,
            module_name,
            fn_name,
            fn_type,
            resources,
            env_name,
            env_vars,
            run_name,
            args,
            kwargs,
        )
        return res

    # TODO [DG] test this properly
    # def debug(self, redirect_logging=False, timeout=10000, *args, **kwargs):
    #     """Run the Function in debug mode. This will run the Function through a tunnel interpreter, which
    #     allows the use of breakpoints and other debugging tools, like rh.ipython().
    #     FYI, alternative ideas from Ray: https://github.com/ray-project/ray/issues/17197
    #     FYI, alternative Modal folks shared: https://github.com/modal-labs/modal-client/pull/32
    #     """
    #     from paramiko import AutoAddPolicy
    #
    #     # Importing this here because they're heavy
    #     from plumbum.machines.paramiko_machine import ParamikoMachine
    #     from rpyc.utils.classic import redirected_stdio
    #     from rpyc.utils.zerodeploy import DeployedServer
    #
    #     creds = self.system.ssh_creds()
    #     ssh_client = ParamikoMachine(
    #         self.system.address,
    #         user=creds["ssh_user"],
    #         keyfile=str(Path(creds["ssh_private_key"]).expanduser()),
    #         missing_host_policy=AutoAddPolicy(),
    #     )
    #     server = DeployedServer(
    #         ssh_client, server_class="rpyc.utils.server.ForkingServer"
    #     )
    #     conn = server.classic_connect()
    #
    #     if redirect_logging:
    #         rlogger = conn.modules.logging.getLogger()
    #         rlogger.parent = logging.getLogger()
    #
    #     conn._config[
    #         "sync_request_timeout"
    #     ] = timeout  # seconds. May need to be longer for real debugging.
    #     conn._config["allow_public_attrs"] = True
    #     conn._config["allow_pickle"] = True
    #     # This assumes the code is already synced over to the remote container
    #     remote_fn = getattr(conn.modules[self.fn.__module__], self.fn.__name__)
    #
    #     with redirected_stdio(conn):
    #         res = remote_fn(*args, **kwargs)
    #     conn.close()
    #     return res

    @property
    def config_for_rns(self):
        # TODO save Package resource, because fn_pointers are meaningless without the package.

        config = super().config_for_rns

        config.update(
            {
                "system": self._resource_string_for_subconfig(self.system),
                "env": self._resource_string_for_subconfig(self.env),
                "fn_pointers": self.fn_pointers,
                "resources": self.resources,
            }
        )
        return config

    def _save_sub_resources(self):
        if isinstance(self.system, Resource):
            self.system.save()

    # TODO maybe reuse these if we starting putting each function in its own container
    # @staticmethod
    # def run_ssh_cmd_in_cluster(ssh_key, ssh_user, address, cmd, port_fwd=None):
    #     subprocess.run("ssh -tt -o IdentitiesOnly=yes -i "
    #                    f"{ssh_key} {port_fwd or ''}"
    #                    f"{ssh_user}@{address} docker exec -it ray_container /bin/bash -c {cmd}".split(' '))

    def ssh(self):
        """SSH into the system associated with the function.

        Example:
            >>> remote_fn = rh.function(local_fn).to(gpu)
            >>> # SSH into gpu
            >>> remote_fn.ssh()
        """
        if self.system is None:
            raise RuntimeError("System must be specified and up to ssh into a Function")
        self.system.ssh()

    def send_secrets(self, providers: Optional[List[str]] = None):
        """Send secrets to the system.

        Example:
            >>> remote_fn.send_secrets(providers=["aws", "lambda"])
        """
        self.system.sync_secrets(providers=providers)

    def http_url(self, curl_command=False, *args, **kwargs) -> str:
        """
        Return the endpoint needed to run the Function on the remote cluster, or provide the curl command if requested.
        """
        raise NotImplementedError("http_url not yet implemented for Function")

    def notebook(self, persist=False, sync_package_on_close=None, port_forward=8888):
        """Tunnel into and launch notebook from the system."""
        # Roughly trying to follow:
        # https://towardsdatascience.com/using-jupyter-notebook-running-on-a-remote-docker-container-via-ssh-ea2c3ebb9055
        # https://docs.ray.io/en/latest/ray-core/using-ray-with-jupyter.html
        if self.system is None:
            raise RuntimeError("Cannot SSH, running locally")

        tunnel, port_fwd = self.system.ssh_tunnel(
            local_port=port_forward, num_ports_to_try=10
        )
        try:
            install_cmd = "pip install jupyterlab"
            jupyter_cmd = f"jupyter lab --port {port_fwd} --no-browser"
            # port_fwd = '-L localhost:8888:localhost:8888 '  # TOOD may need when we add docker support
            with self.system.pause_autostop():
                self.system.run(commands=[install_cmd, jupyter_cmd], stream_logs=True)

        finally:
            if sync_package_on_close:
                if sync_package_on_close == "./":
                    sync_package_on_close = rh_config.rns_client.locate_working_dir()
                pkg = Package.from_string("local:" + sync_package_on_close)
                self.system._rsync(
                    source=f"~/{pkg.name}", dest=pkg.local_path, up=False
                )
            if not persist:
                tunnel.stop()
                kill_jupyter_cmd = f"jupyter notebook stop {port_fwd}"
                self.system.run(commands=[kill_jupyter_cmd])

    def get_or_call(self, run_name: str = None, *args, **kwargs) -> Any:
        """Check if Run was already completed, and if so return the result.
        If no cached Run is found on the cluster, create a new one and run it synchronously before
        returning its result.

        Args:
            run_name (Optional[str]): Name of a particular run for this function.
                If not provided will use the function's name.
            *args: Arguments to pass to the function for the run (relevant if creating a new run).
            **kwargs: Keyword arguments to pass to the function for the run (relevant if creating a new run).

        Returns:
            Any: Result of the Run

        Example:
            >>> # previously, remote_fn.run(arg1, arg2, run_name="my_async_run")
            >>> remote_fn.get_or_call()
        """
        from runhouse import Run

        run_name = run_name or Run._create_new_run_name(self.name)

        res = self._call_fn_with_ssh_access(
            fn_type="get_or_call", run_name=run_name, args=args, kwargs=kwargs
        )

        return res

    def get_or_run(self, run_name: str = None, *args, **kwargs) -> "Run":
        """Check if Run was already completed. If no cached Run is found on the cluster, create a new one.

        Note: If the Run has already completed, will not trigger a new Run.

        Args:
            run_name (Optional[str]): Name of a particular run for this function.
                If not provided will use the function's name.
            *args: Arguments to pass to the function for the run (relevant if creating a new run).
            **kwargs: Keyword arguments to pass to the function for the run (relevant if creating a new run).

        Returns:
            Run: Run object

        Example:
            >>> # previously, remote_fn.run(arg1, arg2, run_name="my_async_run")
            >>> remote_fn.get_or_call()
        """
        from runhouse import Run

        run_name = run_name or Run._create_new_run_name(self.name)
        if run_name == "latest":
            # TODO [JL]
            raise NotImplementedError("Latest not currently supported")

        completed_run: "Run" = self._call_fn_with_ssh_access(
            fn_type="get_or_run", run_name=run_name, args=args, kwargs=kwargs
        )

        return completed_run

    def keep_warm(
        self,
        autostop_mins=None,
        # TODO regions: List[str] = None,
        # TODO min_replicas: List[int] = None,
        # TODO max_replicas: List[int] = None
    ):
        """Keep the system warm for autostop_mins. If autostop_mins is ``None`` or -1, keep warm indefinitely.

        Example:
            >>> # keep gpu warm for 30 mins
            >>> remote_fn = rh.function(local_fn).to(gpu)
            >>> remote_fn.keep_warm(autostop_mins=30)
        """
        if autostop_mins is None:
            logger.info(f"Keeping {self.name} indefinitely warm")
            # keep indefinitely warm if user doesn't specify
            autostop_mins = -1
        self.system.keep_warm(autostop_mins=autostop_mins)

    @staticmethod
    def _handle_nb_fn(fn, fn_pointers, serialize_notebook_fn, name):
        """Handle the case where the user passes in a notebook function"""
        if serialize_notebook_fn:
            # This will all be cloudpickled by the RPC client and unpickled by the RPC server
            # Note that this means the function cannot be saved, and it's better that way because
            # pickling functions is not meant for long term storage. Case in point, this method will be
            # sensitive to differences in minor Python versions between the serializing and deserializing envs.
            return "", "notebook", fn
        else:
            # TODO put this in the current folder instead?
            module_path = Path.cwd() / (f"{name}_fn.py" if name else "sent_fn.py")
            logging.info(
                f"Writing out function function to {str(module_path)}. Please make "
                f"sure the function does not rely on any local variables, "
                f"including imports (which should be moved inside the function body)."
            )
            if not name:
                logging.warning(
                    "You should name Functions that are created in notebooks to avoid naming collisions "
                    "between the modules that are created to hold their functions "
                    '(i.e. "sent_fn.py" errors.'
                )
            source = inspect.getsource(fn).strip()
            with module_path.open("w") as f:
                f.write(source)
            return fn_pointers[0], module_path.stem, fn_pointers[2]
            # from importlib.util import spec_from_file_location, module_from_spec
            # spec = spec_from_file_location(config['name'], str(module_path))
            # module = module_from_spec(spec)
            # spec.loader.exec_module(module)
            # new_fn = getattr(module, fn_pointers[2])
            # fn_pointers = Function._extract_fn_paths(raw_fn=new_fn, reqs=config['reqs'])


def function(
    fn: Optional[Union[str, Callable]] = None,
    name: Optional[str] = None,
    system: Optional[Union[str, Cluster]] = None,
    env: Optional[Union[List[str], Env, str]] = None,
    resources: Optional[dict] = None,
    dryrun: bool = False,
    load_secrets: bool = False,
    serialize_notebook_fn: bool = False,
    # args below are deprecated
    reqs: Optional[List[str]] = None,
    setup_cmds: Optional[List[str]] = None,
):
    """Builds an instance of :class:`Function`.

    Args:
        fn (Optional[str or Callable]): The function to execute on the remote system when the function is called.
        name (Optional[str]): Name of the Function to create or retrieve.
            This can be either from a local config or from the RNS.
        system (Optional[str or Cluster]): Hardware (cluster) on which to execute the Function.
            This can be either the string name of a Cluster object, or a Cluster object.
        env (Optional[List[str] or Env or str]): List of requirements to install on the remote cluster, or path to the
            requirements.txt file, or Env object or string name of an Env object.
        resources (Optional[dict]): Optional number (int) of resources needed to run the Function on the Cluster.
            Keys must be ``num_cpus`` and ``num_gpus``.
        dryrun (bool): Whether to create the Function if it doesn't exist, or load the Function object as a dryrun.
            (Default: ``False``)
        load_secrets (bool): Whether or not to send secrets; only applicable if `dryrun` is set to ``False``.
            (Default: ``False``)
        serialize_notebook_fn (bool): If function is of a notebook setting, whether or not to serialized the function.
            (Default: ``False``)

    Returns:
        Function: The resulting Function object.

    Example:
        >>> import runhouse as rh

        >>> cluster = rh.cluster(name="my_cluster")
        >>> def sum(a, b):
        >>>    return a + b

        >>> summer = rh.function(fn=sum, name="my_func").to(cluster, env=['requirements.txt']).save()

        >>> # using the function
        >>> res = summer(5, 8)  # returns 13

        >>> # Load function from above
        >>> reloaded_function = rh.function(name="my_func")
    """
    if name and not any([fn, system, env, resources]):
        # Try reloading existing function
        return Function.from_name(name, dryrun)

    if setup_cmds:
        warnings.warn(
            "``setup_cmds`` argument has been deprecated. "
            "Please pass in setup commands to rh.Env corresponding to the function instead."
        )
    if reqs is not None:
        warnings.warn(
            "``reqs`` argument has been deprecated. Please use ``env`` instead."
        )
        env = Env(reqs=reqs, setup_cmds=setup_cmds, working_dir="./")
    elif not isinstance(env, Env):
        env = _get_env_from(env) or Env()
        env.working_dir = env.working_dir or "./"

    fn_pointers = None
    if callable(fn):
        fn_pointers = Function._extract_fn_paths(raw_fn=fn, reqs=env.reqs)
        if fn_pointers[1] == "notebook":
            fn_pointers = Function._handle_nb_fn(
                fn,
                fn_pointers=fn_pointers,
                serialize_notebook_fn=serialize_notebook_fn,
                name=fn_pointers[2] or name,
            )
    elif isinstance(fn, str):
        # Url must match a regex of the form
        # 'https://github.com/username/repo_name/blob/branch_name/path/to/file.py:func_name'
        # Use a regex to extract username, repo_name, branch_name, path/to/file.py, and func_name
        pattern = (
            r"https://github\.com/(?P<username>[^/]+)/(?P<repo_name>[^/]+)/blob/"
            r"(?P<branch_name>[^/]+)/(?P<path>[^:]+):(?P<func_name>.+)"
        )
        match = re.match(pattern, fn)

        if match:
            username = match.group("username")
            repo_name = match.group("repo_name")
            branch_name = match.group("branch_name")
            path = match.group("path")
            func_name = match.group("func_name")
        else:
            raise ValueError(
                "fn must be a callable or string of the form "
                '"https://github.com/username/repo_name/blob/branch_name/path/to/file.py:func_name"'
            )
        module_name = Path(path).stem
        relative_path = str(repo_name / Path(path).parent)
        fn_pointers = (relative_path, module_name, func_name)
        # TODO [DG] check if the user already added this in their reqs
        repo_package = git_package(
            git_url=f"https://github.com/{username}/{repo_name}.git",
            revision=branch_name,
        )
        env.reqs.insert(0, repo_package)

    system = _get_cluster_from(system)

    new_function = Function(
        fn_pointers=fn_pointers,
        system=system,
        env=env,
        resources=resources,
        access=Function.DEFAULT_ACCESS,
        name=name,
        dryrun=dryrun,
    )

    if load_secrets and not dryrun:
        new_function.send_secrets()

    return new_function



from typing import List, Optional, Union

from ..utils.hardware import RESERVED_SYSTEM_NAMES

from .cluster import Cluster
from .on_demand_cluster import OnDemandCluster


# Cluster factory method
def cluster(
    name: str,
    ips: List[str] = None,
    ssh_creds: Optional[dict] = None,
    dryrun: bool = False,
    **kwargs,
) -> Union[Cluster, OnDemandCluster]:
    """
    Builds an instance of :class:`Cluster`.

    Args:
        name (str): Name for the cluster, to re-use later on.
        ips (List[str], optional): List of IP addresses for the BYO cluster.
        ssh_creds (dict, optional): Dictionary mapping SSH credentials.
            Example: ``ssh_creds={'ssh_user': '...', 'ssh_private_key':'<path_to_key>'}``
        dryrun (bool): Whether to create the Cluster if it doesn't exist, or load a Cluster object as a dryrun.
            (Default: ``False``)

    Returns:
        Union[Cluster, OnDemandCluster]: The resulting cluster.

    Example:
        >>> import runhouse as rh
        >>> gpu = rh.cluster(ips=['<ip of the cluster>'],
        >>>                  ssh_creds={'ssh_user': '...', 'ssh_private_key':'<path_to_key>'},
        >>>                  name='rh-a10x').save()

        >>> # Load cluster from above
        >>> reloaded_cluster = rh.cluster(name="rh-a10x")
    """
    if name and ips is None and ssh_creds is None and not kwargs:
        # If only the name is provided and dryrun is set to True
        return Cluster.from_name(name, dryrun)

    if name in RESERVED_SYSTEM_NAMES:
        raise ValueError(
            f"Cluster name {name} is a reserved name. Please use a different name which is not one of "
            f"{RESERVED_SYSTEM_NAMES}."
        )

    if "instance_type" in kwargs.keys():
        # Commenting out for now. If two creation paths creates confusion let's push people to use
        # ondemand_cluster() instead.
        # warnings.warn(
        #     "The `cluster` factory is intended to be used for static clusters. "
        #     "If you would like to create an on-demand cluster, please use `rh.ondemand_cluster()` instead."
        # )
        return ondemand_cluster(name=name, **kwargs)

    return Cluster(ips=ips, ssh_creds=ssh_creds, name=name, dryrun=dryrun)


# OnDemandCluster factory method
def ondemand_cluster(
    name: str,
    instance_type: Optional[str] = None,
    num_instances: Optional[int] = None,
    provider: Optional[str] = None,
    autostop_mins: Optional[int] = None,
    use_spot: bool = False,
    image_id: Optional[str] = None,
    region: Optional[str] = None,
    dryrun: bool = False,
) -> OnDemandCluster:
    """
    Builds an instance of :class:`OnDemandCluster`.

    Args:
        name (str): Name for the cluster, to re-use later on.
        instance_type (int, optional): Type of cloud instance to use for the cluster. This could
            be a Runhouse built-in type, or your choice of instance type.
        num_instances (int, optional): Number of instances to use for the cluster.
        provider (str, optional): Cloud provider to use for the cluster.
        autostop_mins (int, optional): Number of minutes to keep the cluster up after inactivity,
            or ``-1`` to keep cluster up indefinitely.
        use_spot (bool, optional): Whether or not to use spot instance.
        image_id (str, optional): Custom image ID for the cluster.
        region (str, optional): The region to use for the cluster.
        dryrun (bool): Whether to create the Cluster if it doesn't exist, or load a Cluster object as a dryrun.
            (Default: ``False``)

    Returns:
        OnDemandCluster: The resulting cluster.

    Example:
        >>> import runhouse as rh
        >>> # On-Demand SkyPilot Cluster (OnDemandCluster)
        >>> gpu = rh.ondemand_cluster(name='rh-4-a100s',
        >>>                  instance_type='A100:4',
        >>>                  provider='gcp',
        >>>                  autostop_mins=-1,
        >>>                  use_spot=True,
        >>>                  image_id='my_ami_string',
        >>>                  region='us-east-1',
        >>>                  ).save()

        >>> # Load cluster from above
        >>> reloaded_cluster = rh.cluster(name="rh-4-a100s")
    """
    if name and not any([instance_type, num_instances, provider, image_id, region]):
        # If only the name is provided and dryrun is set to True
        return Cluster.from_name(name, dryrun)

    if name in RESERVED_SYSTEM_NAMES:
        raise ValueError(
            f"Cluster name {name} is a reserved name. Please use a different name which is not one of "
            f"{RESERVED_SYSTEM_NAMES}."
        )

    return OnDemandCluster(
        instance_type=instance_type,
        provider=provider,
        num_instances=num_instances,
        autostop_mins=autostop_mins,
        use_spot=use_spot,
        image_id=image_id,
        region=region,
        name=name,
        dryrun=dryrun,
    )



import os
import shutil
import unittest

from pathlib import Path

import pytest

import runhouse as rh
import yaml
from runhouse.rns.folders.folder import Folder
from runhouse.rns.packages import Package

pip_reqs = ["torch", "numpy"]

# -------- BASE ENV TESTS ----------- #


@pytest.mark.localtest
def test_create_env_from_name_local():
    env_name = "~/local_env"
    local_env = rh.env(name=env_name, reqs=pip_reqs).save()
    del local_env

    remote_env = rh.env(name=env_name)
    assert set(pip_reqs).issubset(set(remote_env.reqs))


@pytest.mark.rnstest
def test_create_env_from_name_rns():
    env_name = "rns_env"
    env = rh.env(name=env_name, reqs=pip_reqs).save()
    del env

    remote_env = rh.env(name=env_name)
    assert set(pip_reqs).issubset(set(remote_env.reqs))


@pytest.mark.localtest
def test_create_env():
    test_env = rh.env(name="test_env", reqs=pip_reqs)
    assert len(test_env.reqs) >= 2
    assert set(pip_reqs).issubset(set(test_env.reqs))


@pytest.mark.clustertest
def test_to_cluster(cpu_cluster):
    test_env = rh.env(name="test_env", reqs=["transformers"])

    test_env.to(cpu_cluster)
    res = cpu_cluster.run_python(["import transformers"])
    assert res[0][0] == 0  # import was successful

    cpu_cluster.run(["pip uninstall transformers -y"])


@pytest.mark.awstest
@pytest.mark.clustertest
def test_to_fs_to_cluster(cpu_cluster, s3_package):
    cpu_cluster.install_packages(["s3fs"])

    test_env_s3 = rh.env(name="test_env_s3", reqs=["s3fs", "scipy", s3_package]).to(
        "s3"
    )
    for req in test_env_s3.reqs:
        if isinstance(req, Package) and isinstance(req.install_target, Folder):
            assert req.install_target.system == "s3"
            assert req.install_target.exists_in_system()

    folder_name = "test_package"
    test_env_cluster = test_env_s3.to(system=cpu_cluster, path=folder_name, mount=True)
    for req in test_env_cluster.reqs:
        if isinstance(req, Package) and isinstance(req.install_target, Folder):
            assert req.install_target.system == cpu_cluster

    assert "sample_file_0.txt" in cpu_cluster.run([f"ls {folder_name}"])[0][1]
    cpu_cluster.run([f"rm -r {folder_name}"])


@pytest.mark.clustertest
def test_function_to_env(cpu_cluster):
    test_env = rh.env(name="test-env", reqs=["parameterized"]).save()

    def summer(a, b):
        return a + b

    rh.function(summer).to(cpu_cluster, test_env)
    res = cpu_cluster.run_python(["import parameterized"])
    assert res[0][0] == 0

    cpu_cluster.run(["pip uninstall parameterized -y"])
    res = cpu_cluster.run_python(["import parameterized"])
    assert res[0][0] == 1

    rh.function(summer, system=cpu_cluster, env="test-env")
    res = cpu_cluster.run_python(["import parameterized"])
    assert res[0][0] == 0

    cpu_cluster.run(["pip uninstall parameterized -y"])


def _get_env_var_value(env_var):
    import os

    return os.environ[env_var]


@pytest.mark.clustertest
def test_function_env_vars(cpu_cluster):
    test_env_var = "TEST_ENV_VAR"
    test_value = "value"
    test_env = rh.env(name="test-env", env_vars={test_env_var: test_value})

    get_env_var_cpu = rh.function(_get_env_var_value).to(cpu_cluster, test_env)
    res = get_env_var_cpu(test_env_var)

    assert res == test_value


@pytest.mark.clustertest
def test_function_env_vars_file(cpu_cluster):
    env_file = ".env"
    contents = ["# comment", "", "ENV_VAR1=value", "# comment with =", "ENV_VAR2 =val2"]
    with open(env_file, "w") as f:
        for line in contents:
            f.write(line + "\n")

    test_env = rh.env(name="test-env", env_vars=env_file)

    get_env_var_cpu = rh.function(_get_env_var_value).to(cpu_cluster, test_env)
    assert get_env_var_cpu("ENV_VAR1") == "value"
    assert get_env_var_cpu("ENV_VAR2") == "val2"

    os.remove(env_file)
    assert not Path(env_file).exists()


@pytest.mark.clustertest
def test_env_git_reqs(cpu_cluster):
    git_package = rh.GitPackage(
        git_url="https://github.com/huggingface/diffusers.git",
        install_method="pip",
        revision="v0.11.1",
    )
    env = rh.env(reqs=[git_package])
    env.to(cpu_cluster)
    res = cpu_cluster.run(["pip freeze | grep diffusers"])
    assert "diffusers" in res[0][1]
    cpu_cluster.run(["pip uninstall diffusers -y"])


@pytest.mark.clustertest
def test_working_dir(cpu_cluster, tmp_path):
    working_dir = tmp_path / "test_working_dir"
    working_dir.mkdir(exist_ok=True)

    env = rh.env(working_dir=str(working_dir))
    assert str(working_dir) in env.reqs

    env.to(cpu_cluster)
    assert working_dir.name in cpu_cluster.run(["ls"])[0][1]

    cpu_cluster.run([f"rm -r {working_dir.name}"])
    assert working_dir.name not in cpu_cluster.run(["ls"])[0][1]


# -------- CONDA ENV TESTS ----------- #


def _get_conda_env(name="rh-test", python_version="3.10.9"):
    conda_env = {
        "name": name,
        "channels": ["defaults"],
        "dependencies": [
            f"python={python_version}",
        ],
    }
    return conda_env


def _get_conda_python_version(env, system):
    system.up_if_not()
    env.to(system)
    py_version = system.run([f"{env._run_cmd} python --version"])
    return py_version[0][1]


@pytest.mark.localtest
def test_conda_env_from_name_local():
    env_name = "~/local_conda_env"
    conda_env = _get_conda_env()
    local_env = rh.env(name=env_name, conda_env=conda_env).save()
    del local_env

    remote_env = rh.env(name=env_name, dryrun=True)
    assert remote_env.conda_yaml == conda_env
    assert remote_env.env_name == conda_env["name"]


@pytest.mark.rnstest
def test_conda_env_from_name_rns():
    env_name = "rns_conda_env"
    conda_env = _get_conda_env()
    env = rh.env(name=env_name, conda_env=conda_env).save()
    del env

    remote_env = rh.env(name=env_name)
    assert remote_env.conda_yaml == conda_env
    assert remote_env.env_name == conda_env["name"]


@pytest.mark.clustertest
def test_conda_env_path_to_system(cpu_cluster):
    env_name = "from_path"
    python_version = "3.9.16"
    tmp_path = Path.cwd() / "test-env"
    file_path = f"{tmp_path}/{env_name}.yml"
    tmp_path.mkdir(exist_ok=True)
    yaml.dump(
        _get_conda_env(name=env_name, python_version=python_version),
        open(file_path, "w"),
    )
    conda_env = rh.env(conda_env=file_path)
    shutil.rmtree(tmp_path)

    assert python_version in _get_conda_python_version(conda_env, cpu_cluster)


@pytest.mark.clustertest
def test_conda_env_local_to_system(cpu_cluster):
    env_name = "local-env"
    python_version = "3.9.16"
    os.system(f"conda create -n {env_name} -y python=={python_version}")

    conda_env = rh.env(conda_env=env_name)
    assert python_version in _get_conda_python_version(conda_env, cpu_cluster)


@pytest.mark.clustertest
def test_conda_env_dict_to_system(cpu_cluster):
    conda_env = _get_conda_env(python_version="3.9.16")
    test_env = rh.env(name="test_env", conda_env=conda_env)

    assert "3.9.16" in _get_conda_python_version(test_env, cpu_cluster)

    # ray installed successfully
    res = cpu_cluster.run([f"{test_env._run_cmd} python -c 'import ray'"])
    assert res[0][0] == 0


@pytest.mark.clustertest
def test_function_to_conda_env(cpu_cluster):
    conda_env = _get_conda_env(python_version="3.9.16")
    test_env = rh.env(name="test_env", conda_env=conda_env)

    def summer(a, b):
        return a + b

    rh.function(summer).to(cpu_cluster, test_env)
    assert "3.9.16" in _get_conda_python_version(test_env, cpu_cluster)


@pytest.mark.clustertest
def test_conda_additional_reqs(cpu_cluster):
    new_conda_env = _get_conda_env(name="test-add-reqs")
    new_conda_env = rh.env(name="conda_env", reqs=["scipy"], conda_env=new_conda_env)
    new_conda_env.to(cpu_cluster)
    res = cpu_cluster.run([f"{new_conda_env._run_cmd} python -c 'import scipy'"])
    assert res[0][0] == 0  # reqs successfully installed
    cpu_cluster.run([f"{new_conda_env._run_cmd} pip uninstall scipy"])


@pytest.mark.clustertest
def test_conda_git_reqs(cpu_cluster):
    conda_env = _get_conda_env(name="test-add-reqs")
    git_package = rh.GitPackage(
        git_url="https://github.com/huggingface/diffusers.git",
        install_method="pip",
        revision="v0.11.1",
    )
    conda_env = rh.env(conda_env=conda_env, reqs=[git_package])
    conda_env.to(cpu_cluster)
    res = cpu_cluster.run([f"{conda_env._run_cmd} pip freeze | grep diffusers"])
    assert "diffusers" in res[0][1]
    cpu_cluster.run([f"{conda_env._run_cmd} pip uninstall diffusers"])


@pytest.mark.clustertest
def test_conda_env_to_fs_to_cluster(cpu_cluster, s3_package):
    conda_env = _get_conda_env(name="s3-env")
    conda_env_s3 = rh.env(reqs=["s3fs", "scipy", s3_package], conda_env=conda_env).to(
        "s3"
    )
    count = 0
    for req in conda_env_s3.reqs:
        if isinstance(req, Package) and isinstance(req.install_target, Folder):
            assert req.install_target.system == "s3"
            assert req.install_target.exists_in_system()
            count += 1
    assert count >= 1

    folder_name = "test_package"
    count = 0
    conda_env_cluster = conda_env_s3.to(
        system=cpu_cluster, path=folder_name, mount=True
    )
    for req in conda_env_cluster.reqs:
        if isinstance(req, Package) and isinstance(req.install_target, Folder):
            assert req.install_target.system == cpu_cluster
            count += 1
    assert count >= 1

    assert "sample_file_0.txt" in cpu_cluster.run([f"ls {folder_name}"])[0][1]
    assert "s3-env" in cpu_cluster.run(["conda info --envs"])[0][1]

    cpu_cluster.run([f"rm -r {folder_name}"])
    cpu_cluster.run(["conda env remove -n s3-env"])


# -------- CONDA ENV + FUNCTION TESTS ----------- #


def np_summer(a, b):
    import numpy as np

    return int(np.sum([a, b]))


@pytest.mark.clustertest
def test_conda_call_fn(cpu_cluster):
    conda_dict = _get_conda_env(name="c")
    conda_env = rh.env(conda_env=conda_dict, reqs=["pytest", "numpy"])
    fn = rh.function(np_summer).to(system=cpu_cluster, env=conda_env)
    result = fn(1, 4)
    assert result == 5


@pytest.mark.clustertest
def test_conda_map_fn(cpu_cluster):
    conda_dict = _get_conda_env(name="test-map-fn")
    conda_env = rh.env(conda_env=conda_dict, reqs=["pytest", "numpy"])
    map_fn = rh.function(np_summer, system=cpu_cluster, env=conda_env)
    inputs = list(zip(range(5), range(4, 9)))
    results = map_fn.starmap(inputs)
    assert results == [4, 6, 8, 10, 12]


if __name__ == "__main__":
    unittest.main()



from typing import Optional

from .folder import Folder


class AzureFolder(Folder):
    RESOURCE_TYPE = "folder"
    DEFAULT_FS = "azure"

    def __init__(self, dryrun: bool, **kwargs):
        super().__init__(dryrun=dryrun, **kwargs)

    @staticmethod
    def from_config(config: dict, dryrun=False):
        """Load config values into the object."""
        return AzureFolder(**config, dryrun=dryrun)

    def rm(self, contents: list = None, recursive: bool = True):
        """Delete Azure folder along with its contents. Optionally provide a list of folder contents to delete."""
        raise NotImplementedError

    def delete_bucket(self):
        """Delete the Azure bucket."""
        raise NotImplementedError

    def _upload(self, src: str, region: Optional[str] = None):
        """Upload a folder to an Azure bucket."""
        raise NotImplementedError

    def _upload_command(self, src: str, dest: str):
        raise NotImplementedError

    def _download(self, dest):
        """Download a folder from a Azure bucket to local dir."""
        raise NotImplementedError

    def _download_command(self, src, dest):
        raise NotImplementedError

    def _to_cluster(self, dest_cluster, path=None, mount=False):
        raise NotImplementedError

    def _to_local(self, dest_path: str, data_config: dict):
        """Copy a folder from an Azure bucket to local dir."""
        raise NotImplementedError

    def _to_data_store(
        self,
        system: str,
        data_store_path: Optional[str] = None,
        data_config: Optional[dict] = None,
    ):
        """Copy folder from Azure to another remote data store (ex: GCS, S3, Azure)"""
        raise NotImplementedError



import json
import logging
import time

import requests

from runhouse.servers.http.http_utils import handle_response, OutputType, pickle_b64

logger = logging.getLogger(__name__)


class HTTPClient:
    """
    Client for cluster RPCs
    """

    DEFAULT_PORT = 50052
    MAX_MESSAGE_LENGTH = 1 * 1024 * 1024 * 1024  # 1 GB
    CHECK_TIMEOUT_SEC = 5

    def __init__(self, host, port=DEFAULT_PORT):
        self.host = host
        self.port = port

    def request(self, endpoint, req_type="post", data=None, err_str=None, timeout=None):
        req_fn = (
            requests.get
            if req_type == "get"
            else requests.put
            if req_type == "put"
            else requests.delete
            if req_type == "delete"
            else requests.post
        )
        response = req_fn(
            f"http://{self.host}:{self.port}/{endpoint}/",
            json={"data": data},
            timeout=timeout,
        )
        if response.status_code != 200:
            raise ValueError(
                f"Error calling {endpoint} on server: {response.content.decode()}"
            )
        output_type = response.json()["output_type"]
        return handle_response(response.json(), output_type, err_str)

    def check_server(self, cluster_config=None):
        self.request(
            "check",
            req_type="post",
            data=json.dumps(cluster_config, indent=4),
            timeout=self.CHECK_TIMEOUT_SEC,
        )

    def install(self, to_install, env=""):
        self.request(
            "env",
            req_type="post",
            data=pickle_b64((to_install, env)),
            err_str=f"Error installing packages {to_install}",
        )

    def run_module(
        self,
        relative_path,
        module_name,
        fn_name,
        fn_type,
        resources,
        conda_env,
        env_vars,
        run_name,
        args,
        kwargs,
    ):
        """
        Client function to call the rpc for run_module
        """
        # Measure the time it takes to send the message
        module_info = [
            relative_path,
            module_name,
            fn_name,
            fn_type,
            resources,
            conda_env,
            env_vars,
            run_name,
            args,
            kwargs,
        ]
        start = time.time()
        res = self.request(
            "run",
            req_type="post",
            data=pickle_b64(module_info),
            err_str=f"Error inside function {fn_type}",
        )
        end = time.time()
        if fn_type not in ["remote", "get_or_run"]:
            # Printing call time for async runs is not useful
            logging.info(
                f"Time to call remote function: {round(end - start, 2)} seconds"
            )
        return res

    # TODO [DG]: maybe just merge cancel into this so we can get log streaming back as we cancel a job (ditto others)
    def get_object(self, key, stream_logs=False):
        """
        Get a value from the server
        """
        res = requests.get(
            f"http://{self.host}:{self.port}/object/",
            json={"data": pickle_b64((key, stream_logs))},
            stream=True,
        )
        if res.status_code != 200:
            raise ValueError(
                f"Error getting key {key} from server: {res.content.decode()}"
            )
        for responses_json in res.iter_content(chunk_size=None):
            for resp in responses_json.decode().split('{"data":')[1:]:
                resp = json.loads('{"data":' + resp)
                output_type = resp["output_type"]
                result = handle_response(
                    resp, output_type, f"Error running or getting key {key}"
                )
                if output_type not in [OutputType.STDOUT, OutputType.STDERR]:
                    return result

    def put_object(self, key, value):
        self.request(
            "object",
            req_type="put",
            data=pickle_b64((key, value)),
            err_str=f"Error putting object {key}",
        )

    def get_run_object(self, run_name, folder_path):
        run_obj = self.request(
            "run_object",
            req_type="get",
            data=pickle_b64((run_name, folder_path)),
            err_str=f"Error getting Run with name {run_name}",
        )

        return run_obj

    def clear_pins(self, pins=None):
        return self.request(
            "object",
            req_type="delete",
            data=pickle_b64((pins or [])),
            err_str=f"Error clearing pins {pins}",
        )

    def cancel_runs(self, keys, force=False):
        # Note keys can be set to "all" to cancel all runs
        return self.request(
            "cancel",
            req_type="post",
            data=pickle_b64((keys, force)),
            err_str=f"Error cancelling runs {keys}",
        )

    def list_keys(self):
        return self.request("keys", req_type="get")

    def add_secrets(self, secrets):
        failed_providers = self.request(
            "secrets",
            req_type="post",
            data=pickle_b64(secrets),
            err_str="Error sending secrets",
        )
        return failed_providers



import importlib
import logging
import os
import sys
import time
from functools import wraps
from pathlib import Path
from typing import Any, Union

import ray

from ray import cloudpickle as pickle

from runhouse import rh_config
from runhouse.rns.api_utils.utils import resolve_absolute_path
from runhouse.rns.utils.env import _env_vars_from_file

logger = logging.getLogger(__name__)


def call_fn_by_type(
    fn_type,
    fn_name,
    relative_path,
    module_name,
    resources,
    conda_env=None,
    env_vars=None,
    run_name=None,
    args=None,
    kwargs=None,
    serialize_res=True,
):
    from runhouse import Run, RunStatus

    # **NOTE**: No need to pass in the fn callable here! We should be able to get it from the module_name and fn_name

    run_name = run_name if run_name else Run._create_new_run_name(fn_name)

    if fn_type.startswith("get_or"):
        # If Run already exists with this run key return the Run object
        run_obj = _load_existing_run_on_cluster(run_name)
        if run_obj is not None:
            logger.info(f"Found existing Run {run_obj.name}")
            if fn_type == "get_or_run":
                logger.info(
                    f"Found existing Run with name: {run_name}, returning the Run object"
                )
                return run_obj, None, run_name
            elif fn_type == "get_or_call":
                results_path = run_obj._fn_result_path()
                run_status = run_obj.status
                logger.info(f"Found existing Run {run_name} with status: {run_status}")
                try:
                    fn_result = run_obj._load_blob_from_path(path=results_path).fetch()
                    logger.info("Loaded result for Run")
                except FileNotFoundError:
                    # If no result is saved (ex: Run failed) return the stderr or stdout depending on its status
                    logger.info(f"No result found in path: {results_path}")
                    run_output = (
                        run_obj.stderr()
                        if run_status == RunStatus.ERROR
                        else run_obj.stdout()
                    )
                    # Note: API expects serialized result, so pickle before returning
                    fn_result = pickle.dumps(run_output)
                return fn_result, None, run_name
            else:
                raise ValueError(
                    f"Invalid fn_type {fn_type} for an existing Run with name {run_name}"
                )
        else:
            logger.info(f"No existing Run found in file system for {run_name}")
            if fn_type == "get_or_run":
                # No existing run found, continue on with async execution
                fn_type = "remote"
                logger.info(
                    f"No Run found for name {run_name}, creating a new one async"
                )
            elif fn_type == "get_or_call":
                # No existing run found, continue on with synchronous execution
                fn_type = "call"
                logger.info(
                    f"No Run found for name {run_name}, creating a new one and running synchronously"
                )
            else:
                raise ValueError(
                    f"Invalid fn_type {fn_type} for a Run that does not exist with name {run_name}"
                )

    num_gpus = ray.cluster_resources().get("GPU", 0)
    num_cuda_devices = resources.get("num_gpus") or num_gpus

    # We need to add the module_path to the PYTHONPATH because ray runs remotes in a new process
    # We need to set max_calls to make sure ray doesn't cache the remote function and ignore changes to the module
    # See: https://docs.ray.io/en/releases-2.2.0/ray-core/package-ref.html#ray-remote
    module_path = (
        str((Path.home() / relative_path).resolve()) if relative_path else None
    )

    if isinstance(env_vars, str):
        env_vars = _env_vars_from_file(env_vars)
    elif env_vars is None:
        env_vars = {}

    env_vars["PYTHONPATH"] = module_path or ""
    runtime_env = {"env_vars": env_vars}
    if conda_env:
        runtime_env["conda"] = conda_env

    fn_pointers = (module_path, module_name, fn_name)

    logging_wrapped_fn = enable_logging_fn_wrapper(get_fn_from_pointers, run_name)

    ray_fn = ray.remote(
        num_cpus=resources.get("num_cpus") or 0.0001,
        num_gpus=resources.get("num_gpus") or 0.0001 if num_gpus > 0 else None,
        max_calls=len(args) if fn_type in ["starmap"] else 1,
        runtime_env=runtime_env,
    )(logging_wrapped_fn)

    # TODO other possible fn_types: 'batch', 'streaming'
    if fn_type == "starmap":
        obj_ref = [
            ray_fn.remote(fn_pointers, serialize_res, num_cuda_devices, *arg, **kwargs)
            for arg in args
        ]
    elif fn_type in ("queue", "remote", "call", "nested"):
        run_obj = _create_new_run(run_name, fn_name, args, kwargs)
        obj_ref = ray_fn.remote(
            fn_pointers, serialize_res, num_cuda_devices, *args, **kwargs
        )
    elif fn_type == "repeat":
        [num_repeats, args] = args
        obj_ref = [
            ray_fn.remote(fn_pointers, serialize_res, num_cuda_devices, *args, **kwargs)
            for _ in range(num_repeats)
        ]
    else:
        raise ValueError(f"fn_type {fn_type} not recognized")

    if fn_type in ("queue", "remote", "call", "nested"):
        fn_result = _populate_run_with_result(run_obj, fn_type, obj_ref)
        logger.info(f"Result from function execution: {type(fn_result)}")
        res = fn_result
    else:
        res = ray.get(obj_ref)

    # Note: Results by default are serialized
    return res, obj_ref, run_name


def get_fn_from_pointers(fn_pointers, serialize_res, num_gpus, *args, **kwargs):
    (module_path, module_name, fn_name) = fn_pointers

    if module_name == "notebook":
        fn = fn_name  # already unpickled
    else:
        if module_path:
            sys.path.append(module_path)
            logger.info(f"Appending {module_path} to sys.path")

        if module_name in rh_config.obj_store.imported_modules:
            importlib.invalidate_caches()
            rh_config.obj_store.imported_modules[module_name] = importlib.reload(
                rh_config.obj_store.imported_modules[module_name]
            )
            logger.info(f"Reloaded module {module_name}")
        else:
            logger.info(f"Importing module {module_name}")
            rh_config.obj_store.imported_modules[module_name] = importlib.import_module(
                module_name
            )
        fn = getattr(rh_config.obj_store.imported_modules[module_name], fn_name)

    cuda_visible_devices = list(range(int(num_gpus)))
    os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, cuda_visible_devices))

    result = fn(*args, **kwargs)
    if serialize_res:
        return pickle.dumps(result)
    return result


def _stdout_files_for_fn():
    worker_id = ray._private.worker.global_worker.worker_id.hex()

    ray_logs_path = RAY_LOGFILE_PATH
    stdout_files = list(ray_logs_path.glob(f"worker-{worker_id}-*"))

    return stdout_files


def _populate_run_with_result(run_obj, fn_type, obj_ref) -> ["Run", Any]:
    """Execute a function on the cluster. If the ``fn_type`` is ``remote``, then execute asynchronously on another
    thread before returning the Run object. Otherwise execute the function synchronously and return its
    result once completed."""
    run_name = run_obj.name
    logger.info(f"Starting execution of type {fn_type} for Run with name: {run_name}")

    if fn_type == "remote":
        import threading

        logger.info("Running remote function asynchronously on a new thread")

        t = threading.Thread(target=_get_result_from_ray, args=(run_obj, obj_ref))
        t.start()

        # Reload the Run object
        async_run = _load_existing_run_on_cluster(run_name)
        return async_run
    else:
        fn_result, run_status = _get_result_from_ray(run_obj, obj_ref)
        logger.info(
            f"Completed Run for fn type {fn_type} in path: {run_obj.folder.path}"
        )

        # Result should be already serialized
        return fn_result


def _create_new_run(run_name, fn_name, args, kwargs):
    """Create a new Run object and save down relevant config data to its dedicated folder on the cluster."""
    from runhouse.rns.run import Run

    inputs = {"args": args, "kwargs": kwargs}
    logger.info(f"Inputs for Run: {inputs}")

    system = _get_current_system()
    path_to_run = Run._base_cluster_folder_path(run_name)

    logger.info(f"Path to new run: {path_to_run}")
    # Create a new Run object
    new_run = Run(
        name=run_name,
        fn_name=fn_name,
        system=system,
        overwrite=True,
        path=path_to_run,
    )

    logger.info(f"Created new run object from factory: {new_run.run_config}")

    # Save down config for new Run
    new_run._register_new_run()

    # Save down pickled inputs to the function for the Run
    new_run.write(data=pickle.dumps(inputs), path=new_run._fn_inputs_path())

    logger.info(
        f"Finished writing inputs for {run_name} to path: {new_run.folder.path}"
    )

    return new_run


def _get_current_system():
    from runhouse import Folder
    from runhouse.rns.utils.hardware import _current_cluster, _get_cluster_from

    system = _get_cluster_from(
        _current_cluster(key="config") or Folder.DEFAULT_FS, dryrun=True
    )
    return system


def _get_result_from_ray(run_obj, obj_ref):
    from runhouse import RunStatus

    try:
        # Result will not be serialized by default
        result = ray.get(obj_ref)
        logger.info(
            f"Successfully got result of type {type(result)} from Ray object store"
        )

        run_status = RunStatus.COMPLETED

        if not isinstance(result, bytes):
            # Serialize before returning
            result = pickle.dumps(result)

        # Only store result if Run completed successfully
        _save_run_result(result, run_obj)

    except ray.exceptions.RayTaskError as e:
        logger.info(f"Failed to get result from Ray object store: {e}")
        run_status = RunStatus.ERROR
        result = e

        # write traceback to new error file - this will be loaded as the result of the Run
        stderr_path = run_obj._stderr_path
        logger.info(f"Writing error to stderr path: {stderr_path}")
        time.sleep(1)
        run_obj.write(data=str(e).encode(), path=stderr_path)

    finally:
        run_obj._register_fn_run_completion(run_status)

        logger.info(
            f"Registered completed Run {run_obj.name} with status: {run_status}"
        )

        return result, run_status


def _load_existing_run_on_cluster(run_name) -> Union["Run", None]:
    """Load a Run for a given name from the cluster's file system. If the Run is not found returns None."""
    from runhouse import Run, RunType
    from runhouse.rns.folders import folder as folder_factory

    if run_name == "latest":
        # TODO [JL]
        raise NotImplementedError("Latest not currently supported")
    else:
        folder_path = Run._base_cluster_folder_path(run_name)

    folder_path_on_system = resolve_absolute_path(folder_path)
    logger.info(
        f"folder path on system for loading run from file: {folder_path_on_system}"
    )

    system = _get_current_system()
    system_folder = folder_factory(
        path=folder_path_on_system, system=system, dryrun=True
    )

    # Try loading Run from the file system
    run_config = Run._load_run_config(system_folder)
    if not run_config:
        return None

    try:
        # Load the RNS data for this Run - needed to reload the Run object
        rns_config = Run.from_name(run_name, dryrun=True)
        logger.info("Loaded Run metadata from RNS")
    except:
        # No RNS metadata saved for this Run
        logger.info("No RNS metadata found for this Run")
        rns_config = {
            "resource_type": Run.RESOURCE_TYPE,
            "resource_subtype": Run.RESOURCE_TYPE,
            "path": folder_path_on_system,
            "system": system,
            "run_type": RunType.FUNCTION_RUN,
        }

    config = {**run_config, **rns_config}

    return Run.from_config(config, dryrun=True)


def _save_run_result(result: bytes, run_obj):
    """Save the serialized result of the function to the Run's dedicated folder on the cluster"""
    results_path = run_obj._fn_result_path()
    logger.info(f"Writing function result to Run's folder in path: {results_path}")
    run_obj.write(result, path=results_path)


RAY_LOGFILE_PATH = Path("/tmp/ray/session_latest/logs")


def enable_logging_fn_wrapper(fn, run_name):
    @wraps(fn)
    def wrapped_fn(*inner_args, **inner_kwargs):
        """Sets the logfiles for a function to be the logfiles for the current ray worker.
        This is used to stream logs from the worker to the client."""
        logdir = Path.home() / ".rh/logs" / run_name
        logdir.mkdir(exist_ok=True, parents=True)
        # Create simlinks to the ray log files in the run directory
        stdout_files: list = _stdout_files_for_fn()
        for f in stdout_files:
            symlink = logdir / f.name
            if not symlink.exists():
                symlink.symlink_to(f)
        return fn(*inner_args, **inner_kwargs)

    return wrapped_fn



import copy
import logging
from pathlib import Path
from typing import Dict, Optional

from runhouse.rh_config import rns_client
from runhouse.rns.api_utils.utils import generate_uuid
from runhouse.rns.folders import Folder, folder
from runhouse.rns.resource import Resource
from runhouse.rns.utils.hardware import _current_cluster, _get_cluster_from

logger = logging.getLogger(__name__)


class Blob(Resource):
    RESOURCE_TYPE = "blob"
    DEFAULT_FOLDER_PATH = "/runhouse-blob"
    DEFAULT_CACHE_FOLDER = ".cache/runhouse/blobs"

    def __init__(
        self,
        path: Optional[str] = None,
        name: Optional[str] = None,
        system: Optional[str] = Folder.DEFAULT_FS,
        data_config: Optional[Dict] = None,
        dryrun: bool = False,
        **kwargs,
    ):
        """
        Runhouse Blob object

        .. note::
                To build a Blob, please use the factory method :func:`blob`.
        """
        super().__init__(name=name, dryrun=dryrun)
        self._filename = str(Path(path).name) if path else self.name
        # Use factory method so correct subclass for system is returned
        self._folder = folder(
            path=str(Path(path).parent) if path is not None else path,
            system=system,
            data_config=data_config,
            dryrun=dryrun,
        )
        self._cached_data = None

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        blob_config = {
            "path": self.path,  # pair with data source to create the physical URL
            "resource_type": self.RESOURCE_TYPE,
            "system": self._resource_string_for_subconfig(self.system),
        }
        config.update(blob_config)
        return config

    @staticmethod
    def from_config(config: dict, dryrun=False):
        return Blob(**config, dryrun=dryrun)

    @property
    def data(self):
        """Get the blob data."""
        if self._cached_data is not None:
            return self._cached_data
        data = self.fetch()
        return data

    @data.setter
    def data(self, new_data):
        """Update the data blob to new data."""
        self._cached_data = new_data

    @property
    def system(self):
        return self._folder.system

    @system.setter
    def system(self, new_system):
        self._folder.system = new_system

    @property
    def path(self):
        return self._folder.path + "/" + self._filename

    @path.setter
    def path(self, new_path):
        self._folder.path = str(Path(new_path).parent)
        self._filename = str(Path(new_path).name)

    @property
    def data_config(self):
        return self._folder.data_config

    @data_config.setter
    def data_config(self, new_data_config):
        self._folder.data_config = new_data_config

    @property
    def fsspec_url(self):
        return self._folder.fsspec_url + "/" + self._filename

    def open(self, mode: str = "rb"):
        """Get a file-like (OpenFile container object) of the blob data.
        User must close the file, or use this method inside of a with statement.

        Example:
            >>> with my_blob.open(mode="wb") as f:
            >>>     f.write(data)
            >>>
            >>> obj = my_blob.open()
        """
        return self._folder.open(self._filename, mode=mode)

    def to(
        self, system, path: Optional[str] = None, data_config: Optional[dict] = None
    ):
        """Return a copy of the blob on the destination system and path.

        Example:
            >>> local_blob = rh.blob(data)
            >>> s3_blob = blob.to("s3")
            >>> cluster_blob = blob.to(my_cluster)
        """
        new_blob = copy.copy(self)
        new_blob._folder = self._folder.to(
            system=system, path=path, data_config=data_config
        )
        return new_blob

    def fetch(self):
        """Return the data for the user to deserialize.

        Example:
            >>> serialized_data = blob.fetch()
        """
        self._cached_data = self._folder.get(self._filename)
        return self._cached_data

    def _save_sub_resources(self):
        if isinstance(self.system, Resource):
            self.system.save()

    def write(self):
        """Save the underlying blob to its specified fsspec URL.

        Example:
            >>> rh.blob(serialized_data, path="path/to/save").write()
        """
        # TODO figure out default behavior for not overwriting but still saving
        # if not overwrite:
        #     TODO check if data_url is already in use
        #     time = datetime.today().strftime('%Y-%m-%d_%H:%M:%S')
        #     self.data_url = self.data_url + time or time

        # TODO check if self._cached_data is None, and if so, don't just download it to then save it again?
        self._folder.mkdir()
        with self.open(mode="wb") as f:
            if not isinstance(self.data, bytes):
                # Avoid TypeError: a bytes-like object is required
                raise TypeError(
                    f"Cannot save blob with data of type {type(self.data)}, data must be serialized"
                )

            f.write(self.data)

        return self

    def rm(self):
        """Delete the blob and the folder it lives in from the file system.

        Example:
            >>> blob = rh.blob(serialized_data, path="saved/path")
            >>> blob.rm()
        """
        self._folder.rm(contents=[self._filename], recursive=False)

    def exists_in_system(self):
        """Check whether the blob exists in the file system

        Example:
            >>> blob = rh.blob(serialized_data, path="saved/path")
            >>> blob.exists_in_system()
        """
        return self._folder.fsspec_fs.exists(self.fsspec_url)

    # TODO [DG] get rid of this in favor of just "sync_down(path, system)" ?
    def sync_from_cluster(self, cluster, path: Optional[str] = None):
        """Efficiently rsync down a blob from a cluster, into the path of the current Blob object.

        Example:
            >>> remote_blob = rh.blob(serialized_data, system=my_cluster)
            >>> remote_blob.sync_from_cluster()
        """
        if not cluster.address:
            raise ValueError("Cluster must be started before copying data to it.")

        cluster._rsync(source=self.path, dest=path, up=False)


def blob(
    data=None,
    name: Optional[str] = None,
    path: Optional[str] = None,
    system: Optional[str] = None,
    data_config: Optional[Dict] = None,
    dryrun: bool = False,
):
    """Returns a Blob object, which can be used to interact with the resource at the given path

    Args:
        data: Blob data. This should be provided as a serialized object.
        name (Optional[str]): Name to give the blob object, to be reused later on.
        path (Optional[str]): Path (or path) of the blob object.
        system (Optional[str or Cluster]): File system or cluster name. If providing a file system this must be one of:
            [``file``, ``github``, ``sftp``, ``ssh``, ``s3``, ``gs``, ``azure``].
            We are working to add additional file system support.
        data_config (Optional[Dict]): The data config to pass to the underlying fsspec handler.
        dryrun (bool): Whether to create the Blob if it doesn't exist, or load a Blob object as a dryrun.
            (Default: ``False``)

    Returns:
        Blob: The resulting blob.

    Example:
        >>> import runhouse as rh
        >>> import json
        >>> data = json.dumps(list(range(50))
        >>>
        >>> # Remote blob with name and no path (saved to bucket called runhouse/blobs/my-blob)
        >>> rh.blob(name="@/my-blob", data=data, system='s3').write()
        >>>
        >>> # Remote blob with name and path
        >>> rh.blob(name='@/my-blob', path='/runhouse-tests/my_blob.pickle', system='s3').save()
        >>>
        >>> # Local blob with name and path, save to local filesystem
        >>> rh.blob(data=data, path=str(Path.cwd() / "my_blob.pickle")).write()
        >>>
        >>> # Local blob with name and no path (saved to ~/.cache/blobs/my-blob)
        >>> rh.blob(name="~/my-blob", data=data).write().save()

        >>> # Loading a blob
        >>> my_local_blob = rh.blob(name="~/my_blob")
        >>> my_s3_blob = rh.blob(name="@/my_blob")
    """
    if name and not any([data, path, system, data_config]):
        # Try reloading existing blob
        return Blob.from_name(name, dryrun)

    system = _get_cluster_from(
        system or _current_cluster(key="config") or Folder.DEFAULT_FS, dryrun=dryrun
    )

    if path is None:
        blob_name_in_path = (
            f"{generate_uuid()}/{rns_client.resolve_rns_data_resource_name(name)}"
        )

        if system == rns_client.DEFAULT_FS or (
            isinstance(system, Resource) and system.on_this_cluster()
        ):
            # create random path to store in .cache folder of local filesystem
            path = str(
                Path(f"~/{Blob.DEFAULT_CACHE_FOLDER}/{blob_name_in_path}").expanduser()
            )
        else:
            # save to the default bucket
            path = f"{Blob.DEFAULT_FOLDER_PATH}/{blob_name_in_path}"

    new_blob = Blob(
        path=path, system=system, data_config=data_config, name=name, dryrun=dryrun
    )
    new_blob.data = data

    return new_blob



import unittest
from pathlib import Path
from pprint import pprint

import pytest
import runhouse as rh

CTX_MGR_RUN = "my_run_activity"
CLI_RUN_NAME = "my_cli_run"

PATH_TO_CTX_MGR_RUN = f"{rh.Run.LOCAL_RUN_PATH}/{CTX_MGR_RUN}"

RUN_FILES = (
    rh.Run.INPUTS_FILE,
    rh.Run.RESULT_FILE,
    rh.Run.RUN_CONFIG_FILE,
    ".out",
    ".err",
)


# ------------------------- FUNCTION RUN ----------------------------------


@pytest.mark.clustertest
@pytest.mark.runstest
def test_read_fn_stdout(cpu_cluster, submitted_run):
    """Reads the stdout for the Run."""
    fn_run = cpu_cluster.get_run(submitted_run)
    stdout = fn_run.stdout()
    pprint(stdout)
    assert stdout


@pytest.mark.clustertest
@pytest.mark.runstest
def test_load_run_result(cpu_cluster, submitted_run):
    """Load the Run created above directly from the cluster."""
    # Note: Run only exists on the cluster at this point (hasn't yet been saved to RNS).
    func_run = cpu_cluster.get_run(submitted_run)
    assert func_run.result() == 3


@pytest.mark.clustertest
@pytest.mark.runstest
def test_read_fn_run_inputs_and_result(cpu_cluster, submitted_run):
    # Load directly from the cluster
    my_run = cpu_cluster.get_run(submitted_run)
    inputs = my_run.inputs()
    assert inputs == {"args": [1, 2], "kwargs": {}}

    refreshed_run = my_run.refresh()
    assert refreshed_run.status == rh.RunStatus.COMPLETED

    output = refreshed_run.result()
    assert output == 3


@pytest.mark.clustertest
@pytest.mark.runstest
def test_get_or_call_from_cache(summer_func, submitted_run):
    """Cached version of synchronous run - if already completed return the result, otherwise run and wait for
    completion before returning the result."""
    # Note: In this test since we already ran the function, it should return the result without re-running
    run_output = summer_func.get_or_call(run_name=submitted_run)
    assert run_output == 3


@pytest.mark.clustertest
@pytest.mark.runstest
def test_get_or_call_no_cache(summer_func):
    """Cached version of synchronous run - if already completed return the result, otherwise run and wait for
    completion before returning the result."""
    # Note: In this test since we do not have a run with this name, it should first execute the function
    # before returning its result
    run_output = summer_func.get_or_call(run_name="another_sync_run", a=1, b=2)
    assert run_output == 3


@pytest.mark.clustertest
@pytest.mark.runstest
def test_invalid_fn_sync_run(summer_func, cpu_cluster):
    """Test error handling for invalid function Run. The function expects to receive integers but
    does not receive any. An error should be thrown via Ray."""
    import ray

    try:
        summer_func.get_or_call(run_name="invalid_run")
    except ray.exceptions.RayTaskError as e:
        assert (
            str(e.args[0])
            == "summer() missing 2 required positional arguments: 'a' and 'b'"
        )


@pytest.mark.clustertest
@pytest.mark.runstest
def test_invalid_fn_async_run(summer_func):
    """Test error handling for invalid function Run. The function expects to receive integers but
    does not receive any. The Run object returned should have a status of `ERROR`, and the
    result should be its stderr."""
    run_obj = summer_func.get_or_run(run_name="invalid_async_run")

    assert run_obj.refresh().status == rh.RunStatus.ERROR
    assert "summer() missing 2 required positional arguments" in run_obj.result()


@pytest.mark.clustertest
@pytest.mark.runstest
def test_get_fn_status_updates(cpu_cluster, slow_func):
    """Run a function that takes a long time to run, confirming that its status changes as we refresh the Run"""
    async_run = slow_func.run(run_name="my_slow_async_run", a=1, b=2)

    assert isinstance(async_run, rh.Run)

    assert async_run.status == rh.RunStatus.RUNNING

    while async_run.refresh().status != rh.RunStatus.COMPLETED:
        # ... do something else while we wait for the run to finish
        pass

    assert async_run.refresh().status == rh.RunStatus.COMPLETED


@unittest.skip("Not implemented yet.")
@pytest.mark.clustertest
@pytest.mark.runstest
def test_get_or_call_latest(summer_func):
    """Cached version of synchronous run - if already completed return the result, otherwise run and wait for
    completion before returning the result."""
    # Note: In this test since we are providing a name of "latest", it should return the latest cached version
    run_output = summer_func.get_or_call("latest")

    assert run_output == 3


@unittest.skip("Not implemented yet.")
@pytest.mark.clustertest
@pytest.mark.runstest
def test_send_run_to_system_on_completion(summer_func, submitted_async_run):
    # Only once the run actually finishes do we send to S3
    async_run = summer_func.run(run_name=submitted_async_run, a=1, b=2).to(
        "s3", on_completion=True
    )

    assert isinstance(async_run, rh.Run)


@pytest.mark.clustertest
@pytest.mark.runstest
def test_run_refresh(slow_func):
    async_run = slow_func.get_or_run(run_name="async_get_or_run", a=1, b=2)

    while async_run.refresh().status in [
        rh.RunStatus.RUNNING,
        rh.RunStatus.NOT_STARTED,
    ]:
        # do stuff .....
        pass

    assert async_run.refresh().status == rh.RunStatus.COMPLETED


@pytest.mark.clustertest
@pytest.mark.runstest
def test_get_async_run_result(summer_func, submitted_async_run):
    """Read the results from an async run."""
    async_run = summer_func.get_or_run(run_name=submitted_async_run)
    assert isinstance(async_run, rh.Run)
    assert async_run.result() == 3


@pytest.mark.clustertest
@pytest.mark.runstest
def test_get_or_run_no_cache(summer_func):
    """Execute function async on the cluster. If a run already exists, do not re-run. Returns a Run object."""
    # Note: In this test since no Run exists with this name, will trigger the function async on the cluster and in the
    # meantime return a Run object.
    async_run = summer_func.get_or_run(run_name="new_async_run", a=1, b=2)
    assert isinstance(async_run, rh.Run)

    run_result = async_run.result()
    assert run_result == 3


@unittest.skip("Not implemented yet.")
@pytest.mark.clustertest
@pytest.mark.runstest
def test_get_or_run_latest(summer_func):
    """Execute function async on the cluster. If a run already exists, do not re-run. Returns a Run object."""
    # Note: In this test since we are providing "latest", will return the latest cached version.
    async_run = summer_func.get_or_run(run_name="latest")
    assert isinstance(async_run, rh.Run)


@pytest.mark.clustertest
@pytest.mark.runstest
def test_delete_async_run_from_system(cpu_cluster, submitted_async_run):
    # Load the run from the cluster and delete its dedicated folder
    async_run = cpu_cluster.get_run(submitted_async_run)
    async_run.folder.rm()
    assert not async_run.folder.exists_in_system()


@pytest.mark.clustertest
@pytest.mark.rnstest
@pytest.mark.runstest
def test_save_fn_run_to_rns(cpu_cluster, submitted_run):
    """Saves run config to RNS"""
    # Load run that lives on the cluster
    func_run = cpu_cluster.get_run(submitted_run)
    assert func_run

    # Save to RNS
    func_run.save(name=submitted_run)

    # Load from RNS
    loaded_run = rh.run(submitted_run)
    assert rh.exists(loaded_run.name, resource_type=rh.Run.RESOURCE_TYPE)


@pytest.mark.clustertest
@pytest.mark.runstest
def test_create_anon_run_on_cluster(summer_func):
    """Create a new Run without giving it an explicit name."""
    # Note: this will run synchronously and return the result
    res = summer_func(1, 2)
    assert res == 3


@unittest.skip("Not yet implemented.")
@pytest.mark.clustertest
@pytest.mark.runstest
def test_latest_fn_run(summer_func):
    run_output = summer_func.get_or_call(run_str="latest")
    assert run_output == 3


@pytest.mark.clustertest
@pytest.mark.runstest
def test_copy_fn_run_from_cluster_to_local(cpu_cluster, submitted_run):
    my_run = cpu_cluster.get_run(submitted_run)
    my_local_run = my_run.to("here")
    assert my_local_run.folder.exists_in_system()

    # Check that all files were copied
    folder_contents = my_local_run.folder.ls()
    for f in folder_contents:
        file_extension = Path(f).suffix
        file_name = f.split("/")[-1]
        assert file_extension in file_name or file_name in RUN_FILES


@pytest.mark.clustertest
@pytest.mark.awstest
@pytest.mark.runstest
def test_copy_fn_run_from_system_to_s3(cpu_cluster, runs_s3_bucket, submitted_run):
    my_run = cpu_cluster.get_run(submitted_run)
    my_run_on_s3 = my_run.to("s3", path=f"/{runs_s3_bucket}/my_test_run")

    assert my_run_on_s3.folder.exists_in_system()

    # Check that all files were copied
    folder_contents = my_run_on_s3.folder.ls()
    for f in folder_contents:
        file_extension = Path(f).suffix
        file_name = f.split("/")[-1]
        assert file_extension in file_name or file_name in RUN_FILES

    # Delete the run from s3
    my_run_on_s3.folder.rm()
    assert not my_run_on_s3.folder.exists_in_system()


@pytest.mark.rnstest
@pytest.mark.runstest
def test_delete_fn_run_from_rns(submitted_run):
    # Load directly from RNS
    loaded_run = rh.run(name=submitted_run)

    loaded_run.delete_configs()
    assert not rh.exists(name=loaded_run.name, resource_type=rh.Run.RESOURCE_TYPE)


# ------------------------- CLI RUN ------------ ----------------------


@pytest.mark.clustertest
@pytest.mark.runstest
def test_create_cli_python_command_run(cpu_cluster):
    # Run python commands on the specified system. Save the run results to the .rh/logs/<run_name> folder of the system.
    return_codes = cpu_cluster.run_python(
        [
            "import runhouse as rh",
            "import pickle",
            "import logging",
            "local_blob = rh.blob(name='local_blob', data=pickle.dumps(list(range(50)))).write()",
            "logging.info(f'Blob path: {local_blob.path}')",
            "local_blob.rm()",
        ],
        run_name=CLI_RUN_NAME,
    )
    pprint(return_codes)

    assert return_codes[0][0] == 0, "Failed to run python commands"
    assert "Blob path" in return_codes[0][1].strip()


@pytest.mark.clustertest
@pytest.mark.runstest
def test_create_cli_command_run(cpu_cluster):
    """Run CLI command on the specified system.
    Saves the Run locally to the rh/<run_name> folder of the local file system."""
    return_codes = cpu_cluster.run(["python --version"], run_name=CLI_RUN_NAME)

    assert return_codes[0][0] == 0, "Failed to run CLI command"
    assert return_codes[0][1].strip() == "Python 3.10.6"


@pytest.mark.clustertest
@pytest.mark.runstest
def test_send_cli_run_to_cluster(cpu_cluster):
    """Send the CLI based Run which was initially saved on the local file system to the cpu cluster."""
    # Load the run from the local file system
    loaded_run = rh.run(
        name=CLI_RUN_NAME, path=f"{rh.Run.LOCAL_RUN_PATH}/{CLI_RUN_NAME}"
    )
    assert loaded_run.refresh().status == rh.RunStatus.COMPLETED
    assert loaded_run.stdout() == "Python 3.10.6"

    # Save to default path on the cluster (~/.rh/logs/<run_name>)
    cluster_run = loaded_run.to(
        cpu_cluster, path=rh.Run._base_cluster_folder_path(name=CLI_RUN_NAME)
    )

    assert cluster_run.folder.exists_in_system()
    assert isinstance(cluster_run.folder.system, rh.Cluster)


@pytest.mark.clustertest
@pytest.mark.runstest
def test_load_cli_command_run_from_cluster(cpu_cluster):
    # At this point the Run exists locally and on the cluster (hasn't yet been saved to RNS).
    # Load from the cluster
    cli_run = cpu_cluster.get_run(CLI_RUN_NAME)
    assert isinstance(cli_run, rh.Run)


@pytest.mark.clustertest
@pytest.mark.rnstest
@pytest.mark.runstest
def test_save_cli_run_to_rns(cpu_cluster):
    # Load the run from the cluster
    cli_run = cpu_cluster.get_run(CLI_RUN_NAME)

    # Save to RNS
    cli_run.save(name=CLI_RUN_NAME)

    # Confirm Run now lives in RNS
    loaded_run = rh.run(CLI_RUN_NAME)
    assert loaded_run


@pytest.mark.clustertest
@pytest.mark.rnstest
@pytest.mark.runstest
def test_read_cli_command_stdout_from_cluster(cpu_cluster):
    # Read the stdout from the cluster
    cli_run = cpu_cluster.get_run(CLI_RUN_NAME)
    cli_stdout = cli_run.stdout()
    assert cli_stdout == "Python 3.10.6"


@pytest.mark.clustertest
@pytest.mark.runstest
def test_delete_cli_run_from_local_filesystem():
    """Delete the config where it was initially saved (in the local ``rh`` folder of the working directory)"""
    # Load the run from the local file system
    cli_run = rh.run(CLI_RUN_NAME, system=rh.Folder.DEFAULT_FS)
    cli_run.folder.rm()

    assert not cli_run.folder.exists_in_system()


@pytest.mark.clustertest
@pytest.mark.runstest
def test_delete_cli_run_from_cluster(cpu_cluster):
    """Delete the config where it was copied to (in the ``~/.rh/logs/<run_name>`` folder of the cluster)"""
    cli_run = cpu_cluster.get_run(CLI_RUN_NAME)
    assert cli_run, f"Failed to load run {CLI_RUN_NAME} from cluster"

    # Update the Run's folder to point to the cluster instead of the local file system
    cli_run.folder.system = cpu_cluster
    cli_run.folder.path = rh.Run._base_cluster_folder_path(name=CLI_RUN_NAME)
    assert cli_run.folder.exists_in_system()

    cli_run.folder.rm()
    assert not cli_run.folder.exists_in_system()

    cli_run = cpu_cluster.get_run(CLI_RUN_NAME)
    assert cli_run is None, f"Failed to delete {cli_run} on cluster"


@pytest.mark.rnstest
@pytest.mark.runstest
def test_delete_cli_run_from_rns():
    # Load from RNS
    loaded_run = rh.run(CLI_RUN_NAME)
    loaded_run.delete_configs()
    assert not rh.exists(name=loaded_run.name, resource_type=rh.Run.RESOURCE_TYPE)


# ------------------------- CTX MANAGER RUN ----------------------------------


@pytest.mark.clustertest
@pytest.mark.rnstest
@pytest.mark.runstest
def test_create_local_ctx_manager_run(summer_func, cpu_cluster):
    from runhouse.rh_config import rns_client

    ctx_mgr_func = "my_ctx_mgr_func"

    with rh.run(path=PATH_TO_CTX_MGR_RUN) as r:
        # Add all Runhouse objects loaded or saved in the context manager to the Run's artifact registry
        # (upstream + downstream artifacts)
        summer_func.save(ctx_mgr_func)

        summer_func(1, 2, run_name="my_new_run")

        current_run = summer_func.system.get_run("my_new_run")
        run_res = current_run.result()
        print(f"Run result: {run_res}")

        cluster_config = rh.load(name=cpu_cluster.name, instantiate=False)
        cluster = rh.Cluster.from_config(config=cluster_config, dryrun=True)
        print(f"Cluster loaded: {cluster.name}")

        summer_func.delete_configs()

    r.save(name=CTX_MGR_RUN)

    print(f"Saved Run with name: {r.name} to path: {r.folder.path}")

    # Artifacts include the rns resolved name (ex: "/jlewitt1/rh-cpu")
    assert r.downstream_artifacts == [
        rns_client.resolve_rns_path(ctx_mgr_func),
        rns_client.resolve_rns_path(cpu_cluster.name),
    ]
    assert r.upstream_artifacts == [
        rns_client.resolve_rns_path(cpu_cluster.name),
    ]


@pytest.mark.localtest
@pytest.mark.runstest
def test_load_named_ctx_manager_run():
    # Load from local file system
    ctx_run = rh.run(path=PATH_TO_CTX_MGR_RUN)
    assert ctx_run.folder.exists_in_system()


@pytest.mark.localtest
@pytest.mark.runstest
def test_read_stdout_from_ctx_manager_run():
    # Load from local file system
    ctx_run = rh.run(path=PATH_TO_CTX_MGR_RUN)
    stdout = ctx_run.stdout()
    pprint(stdout)
    assert stdout


@pytest.mark.rnstest
@pytest.mark.runstest
def test_save_ctx_run_to_rns():
    # Load from local file system
    ctx_run = rh.run(path=PATH_TO_CTX_MGR_RUN)
    ctx_run.save()
    assert rh.exists(name=ctx_run.name, resource_type=rh.Run.RESOURCE_TYPE)


@pytest.mark.rnstest
@pytest.mark.runstest
def test_delete_ctx_run_from_rns():
    # Load from RNS
    loaded_run = rh.run(name=CTX_MGR_RUN)
    loaded_run.delete_configs()

    assert not rh.exists(name=loaded_run.name, resource_type=rh.Run.RESOURCE_TYPE)


@pytest.mark.clustertest
@pytest.mark.runstest
def test_delete_ctx_run_from_local_filesystem():
    # Load from local file system
    ctx_run = rh.run(path=PATH_TO_CTX_MGR_RUN)
    ctx_run.folder.rm()
    assert not ctx_run.folder.exists_in_system()


if __name__ == "__main__":
    unittest.main()
